# -*- coding: UTF-8 -*-

import operator
import itertools
from kupfer.core import learn, relevance

def make_rankables(itr, rank=0):
    return [Rankable(str(obj), obj, rank) for obj in itr]

def wrap_rankable(obj, rank=0):
    return Rankable(str(obj), obj, rank)

class Rankable (object):
    """
    Rankable has an object (represented item),
    value (determines rank) and an associated rank
    """
    # To save memory with (really) many Rankables
    __slots__ = ("rank", "value", "object", "aliases")
    def __init__(self, value, obj, rank=0):
        self.rank = rank
        self.value = value
        self.object = obj
        self.aliases = getattr(obj, "name_aliases", ())
    
    def __str__(self):
        return "%.2f: %r, %r" % (self.rank, self.value, self.object)

    def __repr__(self):
        return "<Rankable %s repres %r at %x>" % (self, self.object, id(self))

def bonus_objects(rankables, key):
    """
    rankables: List[Rankable]
    inncrement each for mnemonic score for key.
    """
    key = key.lower()
    get_record_score = learn.get_record_score
    for obj in rankables:
        obj.rank += get_record_score(obj.object, key)

def bonus_actions(rankables, key):
    """
    generator of @rankables that have mnemonics for @key

    Add bonus for mnemonics and rank_adjust

    rank is added to prev rank, all items are yielded"""
    key = key.lower()
    get_record_score = learn.get_record_score
    for obj in rankables:
        obj.rank += get_record_score(obj.object, key)
        obj.rank += obj.object.rank_adjust
        yield obj

def add_rank_objects(rankables, rank):
    """
    rankables: List[Rankable]
    rank: Fixed rank
    """
    for obj in rankables:
        obj.rank += rank

def _score_for_key(query):
    if len(query) == 1:
        return relevance.score_single
    else:
        return relevance.score

def score_objects(rankables, key):
    """
    rankables: List[Rankable]

    Prune rankables that score low for the key.
    """
    key = key.lower()
    _score = _score_for_key(key)
    for rb in rankables:
        # Rank object
        rb.rank = _score(rb.value, key) * 100
    for rb in rankables:
        if rb.rank < 90:
            rank = rb.rank
            for alias in rb.aliases:
                # consider aliases and change rb.value if alias is better
                # aliases rank lower so that value is chosen when close
                arank = _score(alias, key) * 95
                if arank > rank:
                    rank = arank
                    rb.value = alias
            rb.rank = rank
    rankables[:] = [rb for rb in rankables if rb.rank > 10]

def score_actions(rankables, for_leaf):
    """Alternative (rigid) scoring mechanism for objects,
    putting much more weight in rank_adjust
    """
    get_record_score = learn.get_record_score
    for obj in rankables:
        ra = obj.object.rank_adjust
        ra += learn.get_correlation_bonus(obj.object, for_leaf)
        if ra > 0:
            obj.rank = 50 + ra + get_record_score(obj.object)//2
        elif ra == 0:
            obj.rank = get_record_score(obj.object)
        else:
            obj.rank = -50 + ra + get_record_score(obj.object)
        yield obj


def _max_multiple(iterables, key):
    maxval = None
    for iterable in iterables:
        try:
            new_max = max(iterable, key=key)
        except ValueError:
            continue
        if maxval is None:
            maxval = new_max
        else:
            maxval = max(maxval, new_max, key=key)
    return maxval

def find_best_sort(rankables):
    """
    rankables: List[List[Rankable]]

    A special kind of lazy sort:

    simply find the best ranked item and yield it first,
    then if needed continue by sorting the rest.

    Note: this will duplicate the best item.

    Yield rankables in best rank first order
    """

    key = operator.attrgetter("rank")
    maxval = _max_multiple(rankables, key)
    if maxval is None: return
    yield maxval

    yield from sorted(itertools.chain(*rankables), key=key, reverse=True)



__kupfer_name__ = _("Show Text")
__kupfer_actions__ = (
        "ShowText",
        #"LargeType",
        "ShowNotification",
    )
__description__ = _("Display text in a window")
__version__ = ""
__author__ = "US"

from kupfer.objects import Action, Leaf, TextLeaf
from kupfer import icons, uiutils
from kupfer import textutils


class ShowText (Action):
    def __init__(self):
        Action.__init__(self, _("Show Text"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        uiutils.show_text_result(leaf.get_text_representation(),
            title=_("Show Text"), ctx=ctx)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Display text in a window")
    def get_icon_name(self):
        return "format-text-bold"

class LargeType (Action):
    def __init__(self):
        Action.__init__(self, _("Large Type"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        return self.activate_multiple((leaf, ), ctx)

    def activate_multiple(self, objects, ctx):
        all_texts = []
        for obj in objects:
            all_texts.append(obj.get_text_representation())
        uiutils.show_large_type("\n".join(all_texts), ctx)

    def item_types(self):
        yield Leaf

    def valid_for_item(self, obj):
        return hasattr(obj, "get_text_representation")

    def get_description(self):
        return _("Display text in a window")
    def get_gicon(self):
        return icons.ComposedIcon.new("format-text-bold", "zoom-in")
    def get_icon_name(self):
        return "format-text-bold"

class ShowNotification (Action):
    def __init__(self):
        Action.__init__(self, _("Show Notification"))

    def activate(self, leaf):
        title, body = textutils.extract_title_body(leaf.object)
        if body:
            uiutils.show_notification(title, body,
                    icon_name=self.get_icon_name())
        else:
            uiutils.show_notification(title)

    def item_types(self):
        yield TextLeaf

    def get_icon_name(self):
        return "format-text-bold"




import pickle as pickle
import os

from kupfer import config
from kupfer import conspickle
from kupfer import pretty

mnemonics_filename = "mnemonics.pickle"
CORRELATION_KEY = 'kupfer.bonus.correlation'

## this is a harmless default
_default_actions = {
    '<builtin.AppLeaf gnome-terminal>': '<builtin.LaunchAgain>',
    '<builtin.AppLeaf xfce4-terminal>': '<builtin.LaunchAgain>',
}
_register = {}
_favorites = set()


class Mnemonics (object):
    """
    Class to describe a collection of mnemonics
    as well as the total count
    """
    def __init__(self):
        self.mnemonics = dict()
        self.count = 0
    def __repr__(self):
        return "<%s %d %s>" % (self.__class__.__name__, self.count, "".join(["%s: %d, " % (m,c) for m,c in self.mnemonics.items()]))
    def increment(self, mnemonic=None):
        if mnemonic:
            mcount = self.mnemonics.get(mnemonic, 0)
            self.mnemonics[mnemonic] = mcount + 1
        self.count += 1

    def decrement(self):
        """Decrement total count and the least mnemonic"""
        if self.mnemonics:
            key = min(self.mnemonics, key=lambda k: self.mnemonics[k])
            if self.mnemonics[key] <= 1:
                del self.mnemonics[key]
            else:
                self.mnemonics[key] -= 1
        self.count = max(self.count -1, 0)

    def __bool__(self):
        return self.count > 0
    def get_count(self):
        return self.count
    def get_mnemonics(self):
        return self.mnemonics

class Learning (object):
    @classmethod
    def _unpickle_register(cls, pickle_file):
        try:
            pfile = open(pickle_file, "rb")
        except IOError as e:
            return None
        try:
            data = conspickle.ConservativeUnpickler.loads(pfile.read())
            assert isinstance(data, dict), "Stored object not a dict"
            pretty.print_debug(__name__, "Reading from %s" % (pickle_file, ))
        except (pickle.PickleError, Exception) as e:
            data = None
            pretty.print_error(__name__, "Error loading %s: %s" % (pickle_file, e))
        finally:
            pfile.close()
        return data

    @classmethod
    def _pickle_register(self, reg, pickle_file):
        ## Write to tmp then rename over for atomicity
        tmp_pickle_file = "%s.%s" % (pickle_file, os.getpid())
        pretty.print_debug(__name__, "Saving to %s" % (pickle_file, ))
        with open(tmp_pickle_file, "wb") as output:
            output.write(pickle.dumps(reg, pickle.HIGHEST_PROTOCOL))
        os.rename(tmp_pickle_file, pickle_file)
        return True

def record_search_hit(obj, key=""):
    """
    Record that KupferObject @obj was used, with the optional
    search term @key recording
    """
    name = repr(obj)
    if name not in _register:
        _register[name] = Mnemonics()
    _register[name].increment(key)

def get_record_score(obj, key=""):
    """
    Get total score for KupferObject @obj,
    bonus score is given for @key matches
    """
    name = repr(obj)
    fav = 7 * (name in _favorites)
    if name not in _register:
        return fav
    mns = _register[name]
    if not key:
        cnt = mns.get_count()
        return fav + 50 * (1 - 1.0/(cnt + 1))

    stats = mns.get_mnemonics()
    closescr = sum(stats[m] for m in stats if m.startswith(key))
    mnscore = 30 * (1 - 1.0/(closescr + 1))
    exact = stats.get(key, 0)
    mnscore += 50 * (1 - 1.0/(exact + 1))
    return fav + mnscore


def get_correlation_bonus(obj, for_leaf):
    """
    Get the bonus rank for @obj when used with @for_leaf
    """
    if _register.setdefault(CORRELATION_KEY, {}).get(repr(for_leaf)) == repr(obj):
        return 50
    else:
        return 0

def set_correlation(obj, for_leaf):
    """
    Register @obj to get a bonus when used with @for_leaf
    """
    _register.setdefault(CORRELATION_KEY, {})[repr(for_leaf)] = repr(obj)

def _get_mnemonic_items(in_register):
    return [(k,v) for k,v in in_register.items() if k != CORRELATION_KEY]

def get_object_has_affinity(obj):
    """
    Return if @obj has any positive score in the register
    """
    return bool(_register.get(repr(obj)) or
                _register.get(CORRELATION_KEY, {}).get(repr(obj)))

def erase_object_affinity(obj):
    """
    Remove all track of affinity for @obj
    """
    _register.pop(repr(obj), None)
    _register.get(CORRELATION_KEY, {}).pop(repr(obj), None)

def _prune_register():
    """
    Remove items with chance (len/25000)

    Assuming homogenous records (all with score one) we keep:
    x_n+1 := x_n * (1 - chance)

    To this we have to add the expected number of added mnemonics per
    invocation, est. 10, and we can estimate a target number of saved mnemonics.
    """
    import random
    random.seed()
    rand = random.random

    goalitems = 500
    flux = 2.0
    alpha = flux/goalitems**2

    chance = min(0.1, len(_register)*alpha)
    for leaf, mn in _get_mnemonic_items(_register):
        if rand() > chance:
            continue
        mn.decrement()
        if not mn:
            del _register[leaf]

    l = len(_register)
    pretty.print_debug(__name__, "Pruned register (%d mnemonics)" % l)

def load():
    """
    Load learning database
    """
    global _register

    filepath = config.get_config_file(mnemonics_filename)
    if filepath:
        _register = Learning._unpickle_register(filepath)
    if not _register:
        _register = {}
    if CORRELATION_KEY not in _register:
        _register[CORRELATION_KEY] = _default_actions

def save():
    """
    Save the learning record
    """
    if not _register:
        pretty.print_debug(__name__, "Not writing empty register")
        return
    if len(_register) > 100:
        _prune_register()
    filepath = config.save_config_file(mnemonics_filename)
    Learning._pickle_register(_register, filepath)

def add_favorite(obj):
    _favorites.add(repr(obj))

def remove_favorite(obj):
    _favorites.discard(repr(obj))

def is_favorite(obj):
    return repr(obj) in _favorites



import sys
import threading

from gi.repository import GLib

from kupfer import scheduler, pretty

class Task (object):
    """Represent a task that can be done in the background

    The finish_callback received in Task.start(..) must be stored,
    and regardless if the task exits with an error, or completes
    successfully, the callback *must* be called.

    The finish callback must pass the Task instance itself as
    the only and first argument:

        finish_callback(self)
    """
    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return "<%s.%s name=%r>" % (type(self).__module__,
                                    type(self).__name__,
                                    getattr(self, 'name', None))

    def start(self, finish_callback):
        raise NotImplementedError

class ThreadTask (Task):
    """Run in a thread"""
    def __init__(self, name=None):
        Task.__init__(self, name)
        self._finish_callback = None

    def thread_do(self):
        """Override this to run what should be done in the thread"""
        raise NotImplementedError

    def thread_finish(self):
        """This finish function runs in the main thread after thread
        completion, and can be used to communicate with the GUI.
        """
        pass

    def thread_finally(self, exc_info):
        """Always run at thread finish"""
        if exc_info is not None:
            etype, value, tb = exc_info
            raise etype(value).with_traceback(tb)

    def _thread_finally(self, exc_info):
        try:
            self.thread_finally(exc_info)
        finally:
            self._finish_callback(self)

    def _run_thread(self):
        try:
            self.thread_do()
            GLib.idle_add(self.thread_finish)
        except:
            exc_info = sys.exc_info()
        else:
            exc_info = None
        finally:
            GLib.idle_add(self._thread_finally, exc_info)

    def start(self, finish_callback):
        self._finish_callback = finish_callback
        thread = threading.Thread(target=self._run_thread)
        thread.start()


class TaskRunner (pretty.OutputMixin):
    """Run Tasks in the idle Loop"""
    def __init__(self, end_on_finish):
        self.tasks = set()
        self.end_on_finish = end_on_finish
        scheduler.GetScheduler().connect("finish", self._finish_cleanup)

    def _task_finished(self, task):
        self.output_debug("Task finished", task)
        self.tasks.remove(task)

    def add_task(self, task):
        """Register @task to be run"""
        self.tasks.add(task)
        task.start(self._task_finished)

    def _finish_cleanup(self, sched):
        if self.end_on_finish:
            self.tasks.clear()
            return
        if self.tasks:
            self.output_info("Uncompleted tasks:")
            for task in self.tasks:
                self.output_info(task)



import pkgutil
import sys

from kupfer import pretty
from kupfer.core import settings
# import kupfer.icons on demand later

sources_attribute = "__kupfer_sources__"
text_sources_attribute = "__kupfer_text_sources__"
content_decorators_attribute = "__kupfer_contents__"
action_decorators_attribute = "__kupfer_actions__"
action_generators_attribute = "__kupfer_action_generators__"
settings_attribute = "__kupfer_settings__"

initialize_attribute = "initialize_plugin"
finalize_attribute = "finalize_plugin"

info_attributes = [
        "__kupfer_name__",
        "__version__",
        "__description__",
        "__author__",
    ]

class NotEnabledError (Exception):
    "Plugin may not be imported since it is not enabled"

def get_plugin_ids():
    """Enumerate possible plugin ids;
    return a sequence of possible plugin ids, not
    guaranteed to be plugins"""
    from kupfer import plugin

    def is_plugname(plug):
        return plug != "__init__" and not plug.endswith("_support")

    for importer, modname, ispkg in pkgutil.iter_modules(plugin.__path__):
        if is_plugname(modname):
            yield modname

class FakePlugin (object):
    def __init__(self, plugin_id, attributes, exc_info):
        self.is_fake_plugin = True
        self.exc_info = exc_info
        self.__name__ = plugin_id
        vars(self).update(attributes)
    def __repr__(self):
        return "<%s %s>" % (type(self).__name__, self.__name__)

def get_plugin_info():
    """Generator, yields dictionaries of plugin descriptions

    with at least the fields:
    name
    localized_name
    version
    description
    author
    """
    for plugin_name in sorted(get_plugin_ids()):
        try:
            plugin = import_plugin_any(plugin_name)
            if not plugin:
                continue
            plugin = vars(plugin)
        except ImportError as e:
            pretty.print_error(__name__, "import plugin '%s':" % plugin_name, e)
            continue
        except Exception as e:
            pretty.print_error(__name__, "Could not load '%s'" % plugin_name)
            pretty.print_exc(__name__)
            continue
        localized_name = plugin.get("__kupfer_name__", None)
        desc = plugin.get("__description__", "")
        vers = plugin.get("__version__", "")
        author = plugin.get("__author__", "")
        # skip false matches;
        # all plugins have to have @localized_name
        if localized_name is None:
            continue
        yield {
            "name": plugin_name,
            "localized_name": localized_name,
            "version": vers,
            "description": desc or "",
            "author": author,
            "provides": (),
        }

def get_plugin_desc():
    """Return a formatted list of plugins suitable for printing to terminal"""
    import textwrap
    infos = list(get_plugin_info())
    verlen = max(len(r["version"]) for r in infos)
    idlen = max(len(r["name"]) for r in infos)
    maxlen = 78
    left_margin = 2 + idlen + 1 + verlen + 1
    desc = []
    for rec in infos:
        # Wrap the description and align continued lines
        wrapped = textwrap.wrap(rec["description"], maxlen - left_margin)
        description = ("\n" + " "*left_margin).join(wrapped)
        desc.append("  %s %s %s" %
            (
                rec["name"].ljust(idlen),
                rec["version"].ljust(verlen),
                description,
            ))
    return "\n".join(desc)

_imported_plugins = {}
_plugin_hooks = {}

class LoadingError(ImportError):
    pass

def _truncate_source(text, find_attributes):
    found_info_attributes = set(find_attributes)
    lines = []
    for line in text.splitlines():
        lines.append(line)
        if not line.strip():
            continue
        first_word, *_rest = line.split(None, 1)
        if first_word in found_info_attributes:
            found_info_attributes.discard(first_word)
        if first_word in ("from", "import", "class", "def", "if"):
            raise LoadingError(("Could not pre-load plugin: Fields missing: %r. "
                "These fields need to be defined before any other code, including imports.")
                    % (list(found_info_attributes), ))
        if not found_info_attributes:
            break
    return "\n".join(lines)

def _import_plugin_fake(modpath, error=None):
    """
    Return an object that has the plugin info attributes we can rescue
    from a plugin raising on import.

    @error: If applicable, a tuple of exception info
    """
    loader = pkgutil.get_loader(modpath)
    if not loader:
        return None

    code = loader.get_source(modpath)
    if not code:
        return None

    try:
        filename = loader.get_filename(modpath)
    except AttributeError:
        try:
            filename = loader.archive + loader.prefix
        except AttributeError:
            filename = "<%s>" % modpath

    env = {
        "__name__": modpath,
        "__file__": filename,
        "__builtins__": {"_": _}
    }
    code = _truncate_source(code, info_attributes)
    try:
        eval(compile(code, filename, "exec"), env)
    except Exception as exc:
        pretty.print_error(__name__, "When loading", modpath)
        pretty.print_exc(__name__)
    attributes = dict((k, env.get(k)) for k in info_attributes)
    attributes.update((k, env.get(k)) for k in ["__name__", "__file__"])
    return FakePlugin(modpath, attributes, error)

def _import_hook_fake(pathcomps):
    modpath = ".".join(pathcomps)
    return _import_plugin_fake(modpath)

def _import_hook_true(pathcomps):
    """@pathcomps path components to the import"""
    path = ".".join(pathcomps)
    fromlist = pathcomps[-1:]
    try:
        setctl = settings.GetSettingsController()
        if not setctl.get_plugin_enabled(pathcomps[-1]):
            raise NotEnabledError("%s is not enabled" % pathcomps[-1])
        plugin = __import__(path, fromlist=fromlist)
    except ImportError as exc:
        # Try to find a fake plugin if it exists
        plugin = _import_plugin_fake(path, error=sys.exc_info())
        if not plugin:
            raise
        pretty.print_error(__name__, "Could not import plugin '%s': %s" %
                (plugin.__name__, exc))
    else:
        pretty.print_debug(__name__, "Loading %s" % plugin.__name__)
        pretty.print_debug(__name__, "  from %s" % plugin.__file__)
    return plugin

def _import_plugin_true(name):
    """Try to import the plugin from the package, 
    and then from our plugin directories in $DATADIR
    """
    plugin = None
    try:
        plugin = _staged_import(name, _import_hook_true)
    except ImportError:
        # Reraise to send this up
        raise
    except NotEnabledError:
        raise
    except Exception:
        # catch any other error for plugins and write traceback
        import traceback
        traceback.print_exc()
        pretty.print_error(__name__, "Could not import plugin '%s'" % name)
    return plugin

def _staged_import(name, import_hook):
    "Import plugin @name using @import_hook"
    plugin = None
    try:
        plugin = import_hook(_plugin_path(name))
    except ImportError as e:
        if name not in e.args[0]:
            raise
    return plugin


def import_plugin(name):
    if is_plugin_loaded(name):
        return _imported_plugins[name]
    plugin = None
    try:
        plugin = _import_plugin_true(name)
    except NotEnabledError:
        plugin = _staged_import(name, _import_hook_fake)
    finally:
        # store nonexistant plugins as None here
        _imported_plugins[name] = plugin
    return plugin

def import_plugin_any(name):
    if name in _imported_plugins:
        return _imported_plugins[name]
    return _staged_import(name, _import_hook_fake)

def _plugin_path(name):
    return ("kupfer", "plugin", name)


# Plugin Attributes
def get_plugin_attributes(plugin_name, attrs, warn=False):
    """Generator of the attributes named @attrs
    to be found in plugin @plugin_name
    if the plugin is not found, we write an error
    and yield nothing.

    if @warn, we print a warning if a plugin does not have
    a requested attribute
    """
    try:
        plugin = import_plugin(plugin_name)
    except ImportError as e:
        pretty.print_info(__name__, "Skipping plugin %s: %s" % (plugin_name, e))
        return
    for attr in attrs:
        try:
            obj = getattr(plugin, attr)
        except AttributeError as e:
            if warn:
                pretty.print_info(__name__, "Plugin %s: %s" % (plugin_name, e))
            yield None
        else:
            yield obj

def get_plugin_attribute(plugin_name, attr):
    """Get single plugin attribute"""
    attrs = tuple(get_plugin_attributes(plugin_name, (attr,)))
    obj, = (attrs if attrs else (None, ))
    return obj

def load_plugin_sources(plugin_name, attr=sources_attribute, instantiate=True):
    sources = get_plugin_attribute(plugin_name, attr)
    if not sources:
        return
    for source in get_plugin_attributes(plugin_name, sources, warn=True):
        if source:
            if instantiate:
                yield source()
            else:
                yield source
        else:
            pretty.print_info(__name__, "Source not found for %s" % plugin_name)


# Plugin Initialization & Error
def is_plugin_loaded(plugin_name):
    return (plugin_name in _imported_plugins and
            not getattr(_imported_plugins[plugin_name], "is_fake_plugin", None))

def _loader_hook(modpath):
    modname = ".".join(modpath)
    loader = pkgutil.find_loader(modname)
    if not loader:
        raise ImportError("No loader found for %s" % modname)
    if not loader.is_package(modname):
        raise ImportError("Is not a package")
    return loader

PLUGIN_ICON_FILE = "icon-list"
icons = None

def _load_icons(plugin_name):
    global icons
    if icons is None:
        from kupfer import icons

    try:
        _loader = _staged_import(plugin_name, _loader_hook)
    except ImportError as exc:
        return
    modname = ".".join(_plugin_path(plugin_name))

    try:
        icon_file = pkgutil.get_data(modname, PLUGIN_ICON_FILE)
    except IOError as exc:
        # icon-list file just missing, let is pass silently
        return

    def get_icon_data(basename):
        return pkgutil.get_data(modname, basename)
    icons.parse_load_icon_list(icon_file, get_icon_data, plugin_name)

def initialize_plugin(plugin_name):
    """Initialize plugin.
    Find settings attribute if defined, and initialize it
    """
    _load_icons(plugin_name)
    settings_dict = get_plugin_attribute(plugin_name, settings_attribute)
    if settings_dict:
        settings_dict.initialize(plugin_name)
    initialize = get_plugin_attribute(plugin_name, initialize_attribute)
    if initialize:
        initialize(plugin_name)
    finalize = get_plugin_attribute(plugin_name, finalize_attribute)
    if finalize:
        register_plugin_unimport_hook(plugin_name, finalize, plugin_name)

def unimport_plugin(plugin_name):
    """Remove @plugin_name from the plugin list and dereference its
    python modules.
    """
    # Run unimport hooks
    if plugin_name in _plugin_hooks:
        try:
            for callback, args in reversed(_plugin_hooks[plugin_name]):
                callback(*args)
        except:
            pretty.print_error(__name__, "Error finalizing", plugin_name)
            pretty.print_exc(__name__)
        del _plugin_hooks[plugin_name]
    del _imported_plugins[plugin_name]
    plugin_module_name = ".".join(_plugin_path(plugin_name))
    pretty.print_debug(__name__, "Dereferencing module", plugin_module_name)
    if plugin_module_name in sys.modules:
        sys.modules.pop(plugin_module_name)
    for mod in list(sys.modules):
        if mod.startswith(plugin_module_name + "."):
            pretty.print_debug(__name__, "Dereferencing module", mod)
            sys.modules.pop(mod)

def register_plugin_unimport_hook(plugin_name, callback, *args):
    if plugin_name not in _imported_plugins:
        raise ValueError("No such plugin %s" % plugin_name)
    _plugin_hooks.setdefault(plugin_name, []).append((callback, args))

def get_plugin_error(plugin_name):
    """
    Return None if plugin is loaded without error, else
    return a tuple of exception information
    """
    try:
        plugin = import_plugin(plugin_name)
        if getattr(plugin, "is_fake_plugin", None):
            return plugin.exc_info
    except ImportError:
        return sys.exc_info()




import os
import re

import gi

from gi.repository import GLib, GObject, Gtk, Gio, Gdk
from gi.repository import Pango
from xdg import BaseDirectory as base
from xdg import DesktopEntry as desktop
from xdg import Exceptions as xdg_e


from kupfer import config, pretty, utils, icons, version
from kupfer import scheduler, kupferstring
from kupfer import kupferui
from kupfer.core import settings, plugins, relevance, sources
from kupfer.ui import keybindings
from kupfer.ui.credentials_dialog import ask_user_credentials
from kupfer.ui import getkey_dialog
from kupfer.ui import accelerators
from kupfer import plugin_support

# index in GtkNotebook
PLUGIN_LIST_PAGE = 2

# List icon pixel size
LIST_ICON_SIZE = 18

# A major HACK
# http://tadeboro.blogspot.com/2009/05/wrapping-adn-resizing-gtklabel.html
def _cb_allocate(label, allocation, maxwid):
    if maxwid == -1:
        maxwid = 300
    label.set_size_request(min(maxwid, allocation.width), -1)
    pass

def wrapped_label(text=None, maxwid=-1):
    label = Gtk.Label.new(text)
    label.set_line_wrap(True)
    label.connect("size-allocate", _cb_allocate, maxwid)
    return label

def kobject_should_show(obj):
    try:
        leaf_repr = obj.get_leaf_repr()
    except AttributeError:
        pass
    else:
        if leaf_repr is None:
            return True
        if hasattr(leaf_repr, "is_valid") and not leaf_repr.is_valid():
            return False
    return True

def set_combobox(value, combobox):
    """
    Set activate the alternative in the combobox with text value
    """
    value = str(value)
    col = combobox.get_entry_text_column()
    combobox_store = combobox.get_model()
    for row in combobox.get_model():
        if row[col] == value:
            combobox.set_active_iter(row.iter)
            return


class PreferencesWindowController (pretty.OutputMixin):

    KEYBINDING_NAMES = {
        # TRANS: Names of global keyboard shortcuts
        'keybinding': _("Show Main Interface"),
        'magickeybinding': _("Show with Selection"),
    }

    KEYBINDING_TARGETS = {
        "keybinding": keybindings.KEYBINDING_DEFAULT,
        "magickeybinding": keybindings.KEYBINDING_MAGIC,
    }


    def __init__(self):
        """Load ui from data file"""
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)
        ui_file = config.get_data_file("preferences.ui")

        if ui_file:
            builder.add_from_file(ui_file)
        else:
            self.window = None
            return
        self.window = builder.get_object("preferenceswindow")
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.connect("delete-event", self._close_window)
        self.pluglist_parent = builder.get_object("plugin_list_parent")
        self.dirlist_parent = builder.get_object("directory_list_parent")
        self.plugin_about_parent = builder.get_object("plugin_about_parent")
        self.preferences_notebook = builder.get_object("preferences_notebook")

        self.buttonremovedirectory = builder.get_object("buttonremovedirectory")
        checkautostart = builder.get_object("checkautostart")
        checkstatusicon_gtk = builder.get_object("checkstatusicon_gtk")
        checkstatusicon_ai = builder.get_object("checkstatusicon_ai")
        combo_icons_large_size = builder.get_object("icons_large_size")
        combo_icons_small_size = builder.get_object("icons_small_size")
        checkusecommandkeys = builder.get_object("checkusecommandkeys")
        radio_actionaccelalt = builder.get_object("radio_actionaccelalt")
        radio_actionaccelctrl = builder.get_object("radio_actionaccelctrl")
        self.entry_plugins_filter = builder.get_object('entry_plugins_filter')
        self.keybindings_list_parent = builder.get_object('keybindings_list_parent')
        self.gkeybindings_list_parent = builder.get_object('gkeybindings_list_parent')
        source_list_parent = builder.get_object("source_list_parent")
        button_reset_keys = builder.get_object("button_reset_keys")
        self.sources_list_ctrl = SourceListController(source_list_parent)

        setctl = settings.GetSettingsController()
        checkautostart.set_active(self._get_should_autostart())
        checkstatusicon_gtk.set_active(setctl.get_show_status_icon())

        large_icon_size = setctl.get_config_int("Appearance", "icon_large_size")
        small_icon_size = setctl.get_config_int("Appearance", "icon_small_size")

        set_combobox(large_icon_size, combo_icons_large_size)
        set_combobox(small_icon_size, combo_icons_small_size)

        if supports_app_indicator():
            checkstatusicon_ai.set_active(setctl.get_show_status_icon_ai())
        else:
            checkstatusicon_ai.set_sensitive(False)

        label = checkstatusicon_gtk.get_label()
        checkstatusicon_gtk.set_label(label + " (GtkStatusIcon)")
        label = checkstatusicon_ai.get_label()
        checkstatusicon_ai.set_label(label + " (AppIndicator)")


        checkusecommandkeys.set_active(setctl.get_use_command_keys())
        radio_actionaccelalt.set_active(setctl.get_action_accelerator_modifer() != 'ctrl')
        radio_actionaccelctrl.set_active(setctl.get_action_accelerator_modifer() == 'ctrl')

        # List store with columns (Name, ID) 
        # Make alternative comboboxes
        terminal_combobox = builder.get_object("terminal_combobox")
        icons_combobox = builder.get_object("icons_combobox")

        def make_combobox_model(combobox):
            combobox_store = Gtk.ListStore(GObject.TYPE_STRING,
                                           GObject.TYPE_STRING)
            combobox.set_model(combobox_store)
            combobox_cell = Gtk.CellRendererText()
            combobox.pack_start(combobox_cell, True)
            combobox.add_attribute(combobox_cell, 'text', 0)

        make_combobox_model(terminal_combobox)
        make_combobox_model(icons_combobox)

        self._update_alternative_combobox('terminal', terminal_combobox)
        self._update_alternative_combobox('icon_renderer', icons_combobox)
        self.terminal_combobox = terminal_combobox
        self.icons_combobox = icons_combobox
        setctl.connect("alternatives-changed", self._on_alternatives_changed)


        # Plugin List
        columns = [
            {"key": "plugin_id", "type": str },
            {"key": "enabled", "type": bool },
            {"key": "icon-name", "type": str },
            {"key": "text", "type": str },
        ]
        # setup plugin list table
        column_types = [c["type"] for c in columns]
        self.columns = [c["key"] for c in columns]
        self.store = Gtk.ListStore.new(column_types)
        self.table = Gtk.TreeView.new_with_model(self.store)
        self.table.set_headers_visible(False)
        self.table.set_property("enable-search", False)
        self.table.connect("cursor-changed", self.plugin_table_cursor_changed)
        self.table.get_selection().set_mode(Gtk.SelectionMode.BROWSE)

        checkcell = Gtk.CellRendererToggle()
        checkcol = Gtk.TreeViewColumn("item", checkcell)
        checkcol.add_attribute(checkcell, "active",
                self.columns.index("enabled"))
        checkcell.connect("toggled", self.on_checkplugin_toggled)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_cell.set_property("height", LIST_ICON_SIZE)
        icon_cell.set_property("width", LIST_ICON_SIZE)

        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "icon-name",
                self.columns.index("icon-name"))

        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("item", cell)
        col.add_attribute(cell, "text", self.columns.index("text"))

        self.table.append_column(checkcol)
        # hide icon for now
        #self.table.append_column(icon_col)
        self.table.append_column(col)

        self.plugin_list_timer = scheduler.Timer()
        self.plugin_info = utils.locale_sort(plugins.get_plugin_info(),
                key= lambda rec: rec["localized_name"])
        self._refresh_plugin_list()
        self.output_debug("Standard Plugins: %d" % len(self.store))
        self.table.show()
        self.pluglist_parent.add(self.table)

        # Directory List
        self.dir_store = Gtk.ListStore.new([str, Gio.Icon, str])
        self.dir_table = Gtk.TreeView.new_with_model(self.dir_store)
        self.dir_table.set_headers_visible(False)
        self.dir_table.set_property("enable-search", False)
        self.dir_table.connect("cursor-changed", self.dir_table_cursor_changed)
        self.dir_table.get_selection().set_mode(Gtk.SelectionMode.BROWSE)

        icon_cell = Gtk.CellRendererPixbuf()

        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "gicon", 1)

        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("name", cell)
        col.add_attribute(cell, "text", 2)
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.dir_table.append_column(icon_col)
        self.dir_table.append_column(col)
        self.dir_table.show()
        self.dirlist_parent.add(self.dir_table)
        self.read_directory_settings()

        # global keybindings list
        self.keybind_table, self.keybind_store = _create_conf_keys_list()
        self.keybindings_list_parent.add(self.keybind_table)
        self.keybind_table.connect("row-activated", self.on_keybindings_row_activate)
        button_reset_keys.set_sensitive(keybindings.is_available())
        self.keybind_table.set_sensitive(keybindings.is_available())


        # kupfer interface (accelerators) keybindings list 
        self.gkeybind_table, self.gkeybind_store = _create_conf_keys_list()
        self.gkeybindings_list_parent.add(self.gkeybind_table)
        self.gkeybind_table.connect("row-activated",
                self.on_gkeybindings_row_activate)

        # Requires GTK 3.22
        try:
            self.gkeybindings_list_parent.set_propagate_natural_height(True)
        except AttributeError:
            pass

        self._show_keybindings(setctl)
        self._show_gkeybindings(setctl)

        # Connect to signals at the last point
        builder.connect_signals(self)

    def _show_keybindings(self, setctl):
        names = self.KEYBINDING_NAMES
        self.keybind_store.clear()
        for binding in sorted(names, key=lambda k: names[k]):
            accel = setctl.get_global_keybinding(binding) or ""
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(accel))
            self.keybind_store.append((names[binding], label, binding))

    def _show_gkeybindings(self, setctl):
        names = accelerators.ACCELERATOR_NAMES
        self.gkeybind_store.clear()
        for binding in sorted(names, key=lambda k: names[k]):
            accel = setctl.get_accelerator(binding) or ""
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(accel))
            self.gkeybind_store.append((names[binding], label, binding))

    def read_directory_settings(self):
        setctl = settings.GetSettingsController()
        dirs = setctl.get_directories()
        for d in dirs:
            self.add_directory_model(d, store=False)

    def add_directory_model(self, d, store=False):
        have = list(os.path.normpath(row[0]) for row in self.dir_store)
        if d in have:
            self.output_debug("Ignoring duplicate directory: ", d)
            return
        else:
            have.append(d)

        d = os.path.expanduser(d)
        dispname = utils.get_display_path_for_bytestring(d)
        gicon = icons.get_gicon_for_file(d)
        self.dir_store.append((d, gicon, dispname))

        if store:
            setctl = settings.GetSettingsController()
            setctl.set_directories(have)

    def remove_directory_model(self, rowiter, store=True):
        self.dir_store.remove(rowiter)
        if store:
            have = list(os.path.normpath(row[0]) for row in self.dir_store)
            setctl = settings.GetSettingsController()
            setctl.set_directories(have)

    def on_preferenceswindow_key_press_event(self, widget, event):
        if event.keyval == Gdk.keyval_from_name("Escape"):
            self.hide()
            return True

    def on_checkstatusicon_gtk_toggled(self, widget):
        setctl = settings.GetSettingsController()
        setctl.set_show_status_icon(widget.get_active())

    def on_checkstatusicon_ai_toggled(self, widget):
        setctl = settings.GetSettingsController()
        setctl.set_show_status_icon_ai(widget.get_active())

    def _get_should_autostart(self):
        KUPFER_DESKTOP = "kupfer.desktop"
        AUTOSTART_KEY = "X-GNOME-Autostart-enabled"
        HIDDEN_KEY = "Hidden"
        autostart_dir = base.save_config_path("autostart")
        autostart_file = os.path.join(autostart_dir, KUPFER_DESKTOP)
        if not os.path.exists(autostart_file):
            return False
        try:
            dfile = desktop.DesktopEntry(autostart_file)
        except xdg_e.ParsingError as exception:
            pretty.print_error(__name__, exception)
            return False
        return ((dfile.hasKey(AUTOSTART_KEY) and
                 dfile.get(AUTOSTART_KEY, type="boolean"))
                and (not dfile.hasKey(HIDDEN_KEY) or
                     not dfile.get(HIDDEN_KEY, type="boolean")))

    def on_checkautostart_toggled(self, widget):
        KUPFER_DESKTOP = "kupfer.desktop"
        AUTOSTART_KEY = "X-GNOME-Autostart-enabled"
        HIDDEN_KEY = "Hidden"
        autostart_dir = base.save_config_path("autostart")
        autostart_file = os.path.join(autostart_dir, KUPFER_DESKTOP)
        if not os.path.exists(autostart_file):
            desktop_files = list(base.load_data_paths("applications",
                KUPFER_DESKTOP))
            if not desktop_files:
                self.output_error("Installed kupfer desktop file not found!")
                return
            desktop_file_path = desktop_files[0]
            # Read installed file and modify it
            try:
                dfile = desktop.DesktopEntry(desktop_file_path)
            except xdg_e.ParsingError as exception:
                pretty.print_error(__name__, exception)
                return
            executable = dfile.getExec()
            ## append no-splash
            if "--no-splash" not in executable:
                executable += " --no-splash"
            dfile.set("Exec", executable)
        else:
            try:
                dfile = desktop.DesktopEntry(autostart_file)
            except xdg_e.ParsingError as exception:
                pretty.print_error(__name__, exception)
                return
        activestr = str(bool(widget.get_active())).lower()
        not_activestr = str(not bool(widget.get_active())).lower()
        self.output_debug("Setting autostart to", activestr)
        dfile.set(AUTOSTART_KEY, activestr)
        dfile.set(HIDDEN_KEY, not_activestr)
        ## remove the format specifiers
        executable = dfile.getExec().replace("%F", "")
        dfile.set("Exec", executable)
        dfile.write(filename=autostart_file)

    def on_entrykeybinding_changed(self, widget):
        pass

    def on_buttonkeybinding_clicked(self, widget):
        keystr = getkey_dialog.ask_for_key(keybindings.bind_key,
                    screen=widget.get_screen(),
                    parent=widget.get_toplevel())
        if keystr:
            self.entrykeybinding.set_text(keystr)
            self.output_debug("Try set keybinding with", keystr)
            keybindings.bind_key(keystr)
            setctl = settings.GetSettingsController()
            setctl.set_keybinding(keystr)

    def on_helpbutton_clicked(self, widget):
        kupferui.show_help()

    def on_closebutton_clicked(self, widget):
        self.hide()

    def _refresh_plugin_list(self, us_filter=None):
        "List plugins that pass text filter @us_filter or list all if None"
        self.store.clear()
        setctl = settings.GetSettingsController()

        if us_filter:
            self.plugin_list_timer.set_ms(300, self._show_focus_topmost_plugin)
        else:
            self.plugin_list_timer.invalidate()

        for info in self.plugin_info:
            plugin_id = info["name"]
            if setctl.get_plugin_is_hidden(plugin_id):
                continue
            enabled = setctl.get_plugin_enabled(plugin_id)
            name = info["localized_name"]
            folded_name = kupferstring.tofolded(name)
            desc = info["description"]
            text = "%s" % name

            if us_filter:
                name_score = relevance.score(name, us_filter)
                fold_name_score = relevance.score(folded_name, us_filter)
                desc_score = relevance.score(desc, us_filter)
                if not name_score and not fold_name_score and desc_score < 0.9:
                    continue

            self.store.append((plugin_id, enabled, "kupfer-object", text))

    def _show_focus_topmost_plugin(self):
        try:
            first_row = next(iter(self.store))
        except StopIteration:
            return
        plugin_id = first_row[0]
        self.show_focus_plugin(plugin_id, 0)

    def on_checkplugin_toggled(self, cell, path):
        checkcol = self.columns.index("enabled")
        plugin_id = self._id_for_table_path(path)
        it = self.store.get_iter(path)
        plugin_is_enabled = not self.store.get_value(it, checkcol)
        self.store.set_value(it, checkcol, plugin_is_enabled)
        setctl = settings.GetSettingsController()
        setctl.set_plugin_enabled(plugin_id, plugin_is_enabled)
        self.plugin_sidebar_update(plugin_id)

    def _id_for_table_path(self, path):
        it = self.store.get_iter(path)
        id_col = self.columns.index("plugin_id")
        plugin_id = self.store.get_value(it, id_col)
        return plugin_id

    def _table_path_for_id(self, id_):
        """
        Find the tree path of plugin @id_
        """
        id_col = self.columns.index("plugin_id")
        for row in self.store:
            plugin_id = row[id_col]
            if plugin_id == id_:
                return row.path
        raise ValueError("No such plugin %s" % id_)


    def _plugin_info_for_id(self, plugin_id):
        for info in self.plugin_info:
            if info["name"] == plugin_id:
                return info
        return None

    def plugin_table_cursor_changed(self, table):
        curpath, curcol = table.get_cursor()
        if not curpath:
            return
        plugin_id = self._id_for_table_path(curpath)
        self.plugin_sidebar_update(plugin_id)

    def plugin_sidebar_update(self, plugin_id):
        about = Gtk.VBox()
        about.set_property("spacing", 15)
        about.set_property("border-width", 5)
        info = self._plugin_info_for_id(plugin_id)
        title_label = Gtk.Label()
        m_localized_name = GLib.markup_escape_text(info["localized_name"])
        title_label.set_markup("<b><big>%s</big></b>" % m_localized_name)
        version, description, author = plugins.get_plugin_attributes(plugin_id,
                ( "__version__", "__description__", "__author__", ))
        about.pack_start(title_label, False, True, 0)
        infobox = Gtk.VBox()
        infobox.set_property("spacing", 3)
        # TRANS: Plugin info fields
        for field, val in zip((_("Description"), _("Author")),
                (description, author)):
            if not val:
                continue
            label = Gtk.Label()
            label.set_alignment(0, 0)
            label.set_markup("<b>%s</b>" % field)
            infobox.pack_start(label, False, True, 0)
            label = wrapped_label()
            label.set_alignment(0, 0)
            label.set_markup("%s" % GLib.markup_escape_text(val))
            label.set_selectable(True)
            infobox.pack_start(label, False, True, 0)
        if version:
            label = wrapped_label()
            label.set_alignment(0, 0)
            m_version = GLib.markup_escape_text(version)
            label.set_markup("<b>%s:</b> %s" % (_("Version"), m_version))
            label.set_selectable(True)
            infobox.pack_start(label, False, True, 0)
        about.pack_start(infobox, False, True, 0)

        # Check for plugin load exception
        exc_info = plugins.get_plugin_error(plugin_id)
        if exc_info is not None:
            etype, error, tb = exc_info
            # TRANS: Error message when Plugin needs a Python module to load
            import_error_localized = _("Python module '%s' is needed") % "\\1"
            import_error_pat = "No module named ([^\s]+)"
            errmsg = str(error)
            if re.match(import_error_pat, errmsg):
                errstr = re.sub(import_error_pat,
                        import_error_localized,
                        errmsg, count=1)
            elif issubclass(etype, ImportError):
                errstr = errmsg
            else:
                import traceback
                errstr = "".join(traceback.format_exception(*exc_info))

            label = wrapped_label()
            label.set_alignment(0, 0)
            label.set_markup("<b>%s</b>\n\n%s" % (
                _("Plugin could not be read due to an error:"),
                GLib.markup_escape_text(errstr),
                ))
            label.set_selectable(True)
            about.pack_start(label, False, True, 0)
        elif not plugins.is_plugin_loaded(plugin_id):
            label = Gtk.Label()
            label.set_alignment(0, 0)
            label.set_text("(%s)" % _("disabled"))
            about.pack_start(label, False, True, 0)

        wid = self._make_plugin_info_widget(plugin_id)
        about.pack_start(wid, False, True, 0)
        psettings_wid = self._make_plugin_settings_widget(plugin_id)
        if psettings_wid:
            about.pack_start(psettings_wid, False, True, 0)

        oldch = self.plugin_about_parent.get_child()
        if oldch:
            self.plugin_about_parent.remove(oldch)
        vp = Gtk.Viewport()
        vp.set_shadow_type(Gtk.ShadowType.NONE)
        vp.add(about)
        self.plugin_about_parent.add(vp)
        self.plugin_about_parent.show_all()

    def _make_plugin_info_widget(self, plugin_id):
        sources, actions, text_sources = \
                plugins.get_plugin_attributes(plugin_id, (
                plugins.sources_attribute,
                plugins.action_decorators_attribute,
                plugins.text_sources_attribute)
                )
        vbox = Gtk.VBox()
        vbox.set_property("spacing", 5)
        setctl = settings.GetSettingsController()
        small_icon_size = setctl.get_config_int("Appearance", "icon_small_size")

        def make_objects_frame(objs, title):
            frame_label = Gtk.Label()
            frame_label.set_markup("<b>%s</b>" %
                                   GLib.markup_escape_text(title))
            frame_label.set_alignment(0, 0)
            objvbox = Gtk.VBox()
            objvbox.pack_start(frame_label, False, True, 0)
            objvbox.set_property("spacing", 3)
            for item in objs:
                plugin_type = plugins.get_plugin_attribute(plugin_id, item)
                if not plugin_type:
                    continue
                hbox = Gtk.HBox()
                hbox.set_property("spacing", 3)
                obj = plugin_type()
                name = str(obj)
                desc = obj.get_description() or ""
                gicon = obj.get_icon()
                im = Gtk.Image()
                im.set_property("gicon", gicon)
                im.set_property("pixel-size", small_icon_size)
                hbox.pack_start(im, False, True, 0)
                m_name = GLib.markup_escape_text(name)
                m_desc = GLib.markup_escape_text(desc)
                name_label = \
                    "%s\n<small>%s</small>" % (m_name, m_desc) if m_desc else \
                    "%s" % (m_name, )
                label = wrapped_label()
                label.set_markup(name_label)
                label.set_xalign(0.)
                hbox.pack_start(label, False, True, 0)
                objvbox.pack_start(hbox, True, True, 0)
                # Display information for application content-sources.
                if not kobject_should_show(obj):
                    continue
                try:
                    leaf_repr = obj.get_leaf_repr()
                except AttributeError:
                    continue
                if leaf_repr is None:
                    continue
                hbox = Gtk.HBox()
                hbox.set_property("spacing", 3)
                gicon = leaf_repr.get_icon()
                im = Gtk.Image()
                im.set_property("gicon", gicon)
                im.set_property("pixel-size", small_icon_size // 2)
                hbox.pack_start(Gtk.Label.new(_("Content of")), False, True, 0)
                hbox.pack_start(im, False, True, 0)
                hbox.pack_start(Gtk.Label.new(str(leaf_repr)), False, True, 0)
                objvbox.pack_start(hbox, True, True, 0)
            return objvbox

        sources = list(sources or ()) + list(text_sources or ())
        if sources:
            # TRANS: Plugin contents header
            swid = make_objects_frame(sources, _("Sources"))
            vbox.pack_start(swid, True, True, 0)
        if actions:
            # TRANS: Plugin contents header
            awid = make_objects_frame(actions, _("Actions"))
            vbox.pack_start(awid, True, True, 0)

        vbox.show_all()
        return vbox

    def _get_plugin_change_callback(self, plugin_id, key, value_type,
            get_attr, no_false_values=False):
        """Callback factory for the plugin parameter configuration"""
        def callback(widget):
            value = getattr(widget, get_attr)()
            if no_false_values and not value:
                return
            setctl = settings.GetSettingsController()
            setctl.set_plugin_config(plugin_id, key, value, value_type)
        return callback

    def _get_plugin_credentials_callback(self, plugin_id, key):
        def callback(widget):
            setctl = settings.GetSettingsController()
            val_type = plugin_support.UserNamePassword
            backend_name = plugin_support.UserNamePassword.get_backend_name()
            if plugin_support.UserNamePassword.is_backend_encrypted():
                information = _("Using encrypted password storage: %s") % backend_name
            else:
                information = _("Using password storage: %s") % backend_name
            upass = setctl.get_plugin_config(plugin_id, key, val_type) \
                    or plugin_support.UserNamePassword()
            user_password = ask_user_credentials(upass.username, upass.password, information)
            if user_password:
                upass.username, upass.password = user_password
                setctl.set_plugin_config(plugin_id, key, upass, val_type)
        return callback

    def _make_plugin_settings_widget(self, plugin_id):
        plugin_settings = plugins.get_plugin_attribute(plugin_id,
                plugins.settings_attribute)
        if not plugin_settings:
            return None

        title_label = Gtk.Label()
        # TRANS: Plugin-specific configuration (header)
        title_label.set_markup("<b>%s</b>" % _("Configuration"))
        title_label.set_alignment(0, 0)

        vbox = Gtk.VBox()
        vbox.pack_start(title_label, False, True, 0)
        #vbox.set_property("spacing", 5)

        plugin_settings_keys = iter(plugin_settings) if plugin_settings else ()
        for setting in plugin_settings_keys:
            typ = plugin_settings.get_value_type(setting)
            alternatives = plugin_settings.get_alternatives(setting)
            tooltip = plugin_settings.get_tooltip(setting)
            wid = None
            hbox = Gtk.HBox()
            hbox.set_property("spacing", 10)
            if tooltip:
                hbox.set_tooltip_text(tooltip)
            label = plugin_settings.get_label(setting)

            if issubclass(typ, plugin_support.UserNamePassword):
                wid = Gtk.Button(label or _("Set username and password"))
                wid.connect("clicked", self._get_plugin_credentials_callback(
                        plugin_id, setting))
                hbox.pack_start(wid, False, True, 0)
                vbox.pack_start(hbox, False, True, 0)
                continue

            label_wid = wrapped_label(label, maxwid=200)
            label_wid.set_xalign(0.)
            if issubclass(typ, str):
                if alternatives:
                    wid = Gtk.ComboBoxText.new()
                    val = plugin_settings[setting]
                    active_index = -1
                    for idx, text in enumerate(alternatives):
                        wid.append_text(text)
                        if text == val:
                            active_index = idx
                    if active_index < 0:
                        wid.prepend_text(val)
                        active_index = 0
                    wid.set_active(active_index)
                    wid.connect("changed", self._get_plugin_change_callback(
                        plugin_id, setting, typ, "get_active_text"))
                else:
                    wid = Gtk.Entry()
                    wid.set_text(plugin_settings[setting])
                    wid.connect("changed", self._get_plugin_change_callback(
                        plugin_id, setting, typ, "get_text",
                        no_false_values=True))
                hbox.pack_start(label_wid, False, True, 0)
                hbox.pack_start(wid, True, True, 0)

            elif issubclass(typ, bool):
                wid = Gtk.CheckButton.new_with_label(label)
                wid.set_active(plugin_settings[setting])
                hbox.pack_start(wid, False, True, 0)
                wid.connect("toggled", self._get_plugin_change_callback(
                    plugin_id, setting, typ, "get_active"))
            elif issubclass(typ, int):
                wid = Gtk.SpinButton()
                wid.set_increments(1, 1)
                wid.set_range(0, 1000)
                wid.set_value(plugin_settings[setting])
                hbox.pack_start(label_wid, False, True, 0)
                hbox.pack_start(wid, False, True, 0)
                wid.connect("changed", self._get_plugin_change_callback(
                    plugin_id, setting, typ, "get_text", no_false_values=True))
            vbox.pack_start(hbox, False, True, 0)

        vbox.show_all()
        return vbox

    def on_buttonadddirectory_clicked(self, widget):
        # TRANS: File Chooser Title
        chooser_dialog = Gtk.FileChooserDialog(title=_("Choose a Directory"),
                action=Gtk.FileChooserAction.SELECT_FOLDER,
                buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT,
                    Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT))
        chooser_dialog.set_select_multiple(True)
        if chooser_dialog.run() == Gtk.ResponseType.ACCEPT:
            for selected_dir in chooser_dialog.get_filenames():
                self.add_directory_model(selected_dir, store=True)
        chooser_dialog.hide()

    def on_buttonremovedirectory_clicked(self, widget):
        curpath, curcol = self.dir_table.get_cursor()
        if not curpath:
            return
        it = self.dir_store.get_iter(curpath)
        self.remove_directory_model(it, store=True)

    def on_entry_plugins_filter_changed(self, widget):
        s_filter = widget.get_text()
        us_filter = kupferstring.tounicode(s_filter).lower()
        self._refresh_plugin_list(us_filter)

    def on_entry_plugins_filter_icon_press(self, entry, icon_pos, event):
        entry.set_text('')

    def on_keybindings_row_activate(self, treeview, path, view_column):
        def bind_key_func(target):
            def bind_key(keystr):
                return keybindings.bind_key(keystr, target)
            return bind_key

        it = self.keybind_store.get_iter(path)
        keybind_id = self.keybind_store.get_value(it, 2)
        setctl = settings.GetSettingsController()
        curr_key = setctl.get_global_keybinding(keybind_id)
        bind_func = bind_key_func(self.KEYBINDING_TARGETS[keybind_id])
        keystr = getkey_dialog.ask_for_key(bind_func, curr_key,
                                screen=treeview.get_screen(),
                                parent=treeview.get_toplevel())
        if keystr == '':
            keybindings.bind_key(None, self.KEYBINDING_TARGETS[keybind_id])
            setctl.set_global_keybinding(keybind_id, keystr)
            self.keybind_store.set_value(it, 1, '')
        elif keystr is not None:
            setctl.set_global_keybinding(keybind_id, keystr)
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
            self.keybind_store.set_value(it, 1, label)

    def _is_good_keystr(self, keystr):
        # Reject single letters so you can't bind 'A' etc
        if keystr is None:
            return
        label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
        ulabel = kupferstring.tounicode(label)
        return not (len(ulabel) == 1 and ulabel.isalnum())

    def on_gkeybindings_row_activate(self, treeview, path, view_column):
        it = self.gkeybind_store.get_iter(path)
        keybind_id = self.gkeybind_store.get_value(it, 2)
        setctl = settings.GetSettingsController()
        curr_key = setctl.get_accelerator(keybind_id)
        keystr = getkey_dialog.ask_for_key(self._is_good_keystr,
                previous_key=curr_key, screen=treeview.get_screen(),
                parent=treeview.get_toplevel())
        if keystr is not None:
            setctl.set_accelerator(keybind_id, keystr)
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
            self.gkeybind_store.set_value(it, 1, label)

    def on_button_reset_keys_clicked(self, button):
        if self.ask_user_for_reset_keybinding():
            setctl = settings.GetSettingsController()
            setctl.reset_keybindings()
            self._show_keybindings(setctl)
            # Unbind all before re-binding
            for keybind_id, target in self.KEYBINDING_TARGETS.items():
                keybindings.bind_key(None, target)
            for keybind_id, target in self.KEYBINDING_TARGETS.items():
                keystr = setctl.get_global_keybinding(keybind_id)
                keybindings.bind_key(keystr, target)

    def on_button_reset_gkeys_clicked(self, button):
        if self.ask_user_for_reset_keybinding():
            setctl = settings.GetSettingsController()
            setctl.reset_accelerators()
            self._show_gkeybindings(setctl)

    def on_checkusecommandkeys_toggled(self, widget):
        setctl = settings.GetSettingsController()
        setctl.set_use_command_keys(widget.get_active())

    def on_radio_action_accel_alt(self, widget):
        if widget.get_active():
            setctl = settings.GetSettingsController()
            setctl.set_action_accelerator_modifier('alt')

    def on_radio_action_accel_ctrl(self, widget):
        if widget.get_active():
            setctl = settings.GetSettingsController()
            setctl.set_action_accelerator_modifier('ctrl')

    def dir_table_cursor_changed(self, table):
        curpath, curcol = table.get_cursor()
        if not curpath or not self.dir_store:
            self.buttonremovedirectory.set_sensitive(False)
            return
        self.buttonremovedirectory.set_sensitive(True)

    def on_terminal_combobox_changed(self, widget):
        setctl = settings.GetSettingsController()
        itr = widget.get_active_iter()
        if itr:
            term_id = widget.get_model().get_value(itr, 1)
            setctl.set_preferred_tool('terminal', term_id)

    def on_icons_combobox_changed(self, widget):
        setctl = settings.GetSettingsController()
        itr = widget.get_active_iter()
        if itr:
            term_id = widget.get_model().get_value(itr, 1)
            setctl.set_preferred_tool('icon_renderer', term_id)

    def on_icons_large_size_changed(self, widget):
        setctl = settings.GetSettingsController()
        itr = widget.get_active_iter()
        if itr:
            val = widget.get_active_text()
            setctl.set_large_icon_size(val)

    def on_icons_small_size_changed(self, widget):
        setctl = settings.GetSettingsController()
        itr = widget.get_active_iter()
        if itr:
            val = widget.get_active_text()
            setctl.set_small_icon_size(val)

    def _update_alternative_combobox(self, category_key, combobox):
        """
        Alternatives changed
        """
        combobox_store = combobox.get_model()
        combobox_store.clear()
        setctl = settings.GetSettingsController()
        term_id = setctl.get_preferred_tool(category_key)
        # fill in the available alternatives
        alternatives = utils.locale_sort(
                setctl.get_valid_alternative_ids(category_key), key=lambda t:t[1])
        term_iter = None
        for (id_, name) in alternatives:
            _it = combobox_store.append((name, id_))
            if id_ == term_id:
                term_iter = _it
        # Update selection
        term_iter = term_iter or combobox_store.get_iter_first()
        combobox.set_sensitive(len(combobox_store) > 1)
        if term_iter:
            combobox.set_active_iter(term_iter)


    def _on_alternatives_changed(self, setctl, category_key):
        if category_key == 'terminal':
            self._update_alternative_combobox(category_key,
                    self.terminal_combobox)
        elif category_key == 'icon_renderer':
            self._update_alternative_combobox(category_key,
                    self.icons_combobox)

    def on_preferences_notebook_switch_page(self, notebook, page, page_num):
        ## focus the search box on the plugin tab
        if page_num == PLUGIN_LIST_PAGE:
            GLib.idle_add(self.entry_plugins_filter.grab_focus)

    def show(self, timestamp):
        self.window.present_with_time(timestamp)

    def show_on_screen(self, timestamp, screen):
        self.window.set_screen(screen)
        self.show(timestamp)
        ## focus the search box on the plugin tab
        if self.preferences_notebook.get_current_page() == PLUGIN_LIST_PAGE:
            self.entry_plugins_filter.grab_focus()

    def show_focus_plugin(self, plugin_id, timestamp):
        """
        Open and show information about plugin @plugin_id
        """
        try:
            table_path = self._table_path_for_id(plugin_id)
        except ValueError:
            self.entry_plugins_filter.set_text("")
            self._refresh_plugin_list()
            table_path = self._table_path_for_id(plugin_id)
        self.table.set_cursor(table_path)
        self.table.scroll_to_cell(table_path)
        self.preferences_notebook.set_current_page(PLUGIN_LIST_PAGE)
        self.window.present_with_time(timestamp)

    def hide(self):
        self.window.hide()
    def _close_window(self, *ignored):
        self.hide()
        return True

    def ask_user_for_reset_keybinding(self):
        dlg = Gtk.MessageDialog(self.window, Gtk.DialogFlags.MODAL, Gtk.MessageType.QUESTION)
        dlg.set_markup(_("Reset all shortcuts to default values?"))
        dlg.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CLOSE,
                _('Reset'), Gtk.ResponseType.ACCEPT)
        result = dlg.run() == Gtk.ResponseType.ACCEPT
        dlg.destroy()
        return result


_conf_keys_list_columns = [{"key": "name", "type":str, 'header': _('Command')},
        {"key": "key", "type": str, 'header': _('Shortcut') },
        {"key": "keybinding_id", "type": str, 'header':  None}]
_conf_keys_list_column_types = [c["type"] for c in _conf_keys_list_columns]

def _create_conf_keys_list():
    keybind_store = Gtk.ListStore.new(_conf_keys_list_column_types)
    keybind_table = Gtk.TreeView.new_with_model(keybind_store)
    for idx, col in enumerate(_conf_keys_list_columns):
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(col['header'], renderer, text=idx)
        column.set_visible(col['header'] is not None)
        keybind_table.append_column(column)
    keybind_table.set_property("enable-search", False)
    keybind_table.set_headers_visible(True)
    keybind_table.show()
    return keybind_table, keybind_store


_preferences_window = None

def GetPreferencesWindowController():
    global _preferences_window
    if _preferences_window is None:
        _preferences_window = PreferencesWindowController()
    return _preferences_window

class SourceListController (object):
    def __init__(self, parent_widget):
        columns = [
            {"key": "source", "type": GObject.TYPE_PYOBJECT },
            {"key": "plugin_id", "type": str },
            {"key": "toplevel", "type": bool },
            {"key": "icon", "type": Gio.Icon },
            {"key": "text", "type": str },
        ]
        # setup plugin list table
        column_types = [c["type"] for c in columns]
        self.columns = [c["key"] for c in columns]
        self.store = Gtk.ListStore(*column_types)
        self.table = Gtk.TreeView.new_with_model(self.store)
        self.table.set_headers_visible(False)
        self.table.set_property("enable-search", False)
        #self.table.connect("cursor-changed", self.plugin_table_cursor_changed)
        self.table.get_selection().set_mode(Gtk.SelectionMode.NONE)

        checkcell = Gtk.CellRendererToggle()
        checkcol = Gtk.TreeViewColumn("item", checkcell)
        checkcol.add_attribute(checkcell, "active",
                self.columns.index("toplevel"))
        checkcell.connect("toggled", self.on_checktoplevel_enabled)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_cell.set_property("height", LIST_ICON_SIZE)
        icon_cell.set_property("width", LIST_ICON_SIZE)

        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "gicon",
                self.columns.index("icon"))

        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("item", cell)
        col.add_attribute(cell, "text", self.columns.index("text"))

        self.table.append_column(checkcol)
        self.table.append_column(icon_col)
        self.table.append_column(col)

        self._refresh()
        self.table.show()
        parent_widget.add(self.table)

        setctl = settings.GetSettingsController()
        setctl.connect("plugin-enabled-changed", self._refresh)

    def _refresh(self, *ignored):
        self.store.clear()
        setctl = settings.GetSettingsController()
        sc = sources.GetSourceController()
        srcs = sorted(sc.get_sources(), key=str)

        for src in srcs:
            name = str(src)
            plugin_id = sc.get_plugin_id_for_object(src)
            if not plugin_id or setctl.get_plugin_is_hidden(plugin_id):
                continue
            if not kobject_should_show(src):
                continue
            gicon = src.get_icon()
            toplevel = setctl.get_source_is_toplevel(plugin_id, src)

            self.store.append((src, plugin_id, toplevel, gicon, name))

    def on_checktoplevel_enabled(self, cell, path):
        it = self.store.get_iter(path)
        checkcol = self.columns.index("toplevel")
        idcol = self.columns.index("plugin_id")
        srccol = self.columns.index("source")
        is_toplevel = not self.store.get_value(it, checkcol)
        plugin_id = self.store.get_value(it, idcol)
        src = self.store.get_value(it, srccol)

        sc = sources.GetSourceController()
        sc.set_toplevel(src, is_toplevel)

        setctl = settings.GetSettingsController()
        setctl.set_source_is_toplevel(plugin_id, src, is_toplevel)
        self.store.set_value(it, checkcol, is_toplevel)


def supports_app_indicator():
    try:
        gi.require_version("AppIndicator3", "0.1")
    except ValueError:
        return False
    else:
        return True




from gi.repository import Gtk

from kupfer import version, config, kupferstring

class CredentialsDialogController():
    def __init__(self, username, password, infotext=None):
        """Load ui from data file"""
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)
        ui_file = config.get_data_file("credentials_dialog.ui")
        builder.add_from_file(ui_file)
        builder.connect_signals(self)

        self.window = builder.get_object("credentials_dialog")
        self.entry_user = builder.get_object('entry_username')
        self.entry_pass = builder.get_object('entry_password')
        if infotext:
            hbox_information = builder.get_object('hbox_information')
            label_information = builder.get_object('label_information')
            hbox_information.show()
            label_information.set_text(infotext)

        self.entry_user.set_text(username or '')
        self.entry_pass.set_text(password or '')

    def on_button_ok_clicked(self, widget):
        self.window.response(Gtk.ResponseType.ACCEPT)
        self.window.hide()

    def on_button_cancel_clicked(self, widget):
        self.window.response(Gtk.ResponseType.CANCEL)
        self.window.hide()

    def show(self):
        return self.window.run()

    @property
    def username(self):
        return kupferstring.tounicode(self.entry_user.get_text())

    @property
    def password(self):
        return kupferstring.tounicode(self.entry_pass.get_text())


def ask_user_credentials(user=None, password=None, infotext=None):
    ''' Ask user for username and password.
    
    @user, @password - initial values
    @return:
    (user, password) when user press "change"
    None when user press "cancel" button '''
    dialog = CredentialsDialogController(user, password, infotext)
    result = None
    if dialog.show() == Gtk.ResponseType.ACCEPT:
        result = dialog.username, dialog.password
    return result




import contextlib
import os

from gi.repository import Gtk, Gdk

from kupfer import pretty
from kupfer.ui import keybindings

def gui_context_from_widget(timestamp, widget):
    return GUIEnvironmentContext(timestamp, widget.get_screen())

def gui_context_from_timestamp(timestamp):
    return GUIEnvironmentContext(timestamp, None)

def gui_context_from_keyevent(timestamp, display):
    new_display = GUIEnvironmentContext.ensure_display_open(display)
    screen, x, y, modifiers = new_display.get_pointer()
    return GUIEnvironmentContext(timestamp, screen)

class GUIEnvironmentContext (object):
    """
    Context object for action execution
    in the current GUI context
    """
    _open_displays = set()

    def __init__(self, timestamp, screen=None):
        self._timestamp = timestamp
        self._screen = screen or Gdk.Screen.get_default()

    def __repr__(self):
        return "<%s time=%r screen=%r>" % (
                type(self).__name__,
                self._timestamp,
                self._screen)

    @classmethod
    def ensure_display_open(cls, display):
        """
        Return GdkDisplay for name @display.

        Return default if @display is None.
        """
        return Gdk.DisplayManager.get().get_default_display()
        def norm_name(name):
            "normalize display name"
            if name[-2] == ":":
                return name+".0"
            return name
        dm = Gdk.display_manager_get()
        if display:
            new_display = None
            for disp in dm.list_displays():
                if norm_name(disp.get_name()) == norm_name(display):
                    new_display = disp
                    break
            if new_display is None:
                pretty.print_debug(__name__,
                        "Opening display in ensure_display_open", display)
                new_display = Gdk.Display(display)
        else:
            new_display = Gdk.Display.get_default()
        ## Hold references to all open displays
        cls._open_displays = set(dm.list_displays())
        return new_display

    @classmethod
    def _try_close_unused_displays(cls, screen):
        """@screen is current GdkScreen

        Try to close inactive displays...
        Take all GtkWindow that are hidden, and move to the
        current screen. If no windows remain then we close
        the display, but we never close the default display.
        """
        def debug(*x):
            pretty.print_debug(__name__, *x)
        display = screen.get_display()
        dm = Gdk.DisplayManager.get()
        for disp in list(dm.list_displays()):
            if disp != display and disp != Gdk.Display.get_default():
                debug("Trying to close", disp.get_name())
                open_windows = 0
                for window in Gtk.window_list_toplevels():
                    # find windows on @disp
                    if window.get_screen().get_display() != disp:
                        continue
                    if not window.get_property("visible"):
                        debug("Moving window", window.get_name())
                        debug("Moving", window.get_title())
                        window.set_screen(screen)
                    else:
                        debug("Open window blocks close")
                        open_windows += 1
                if not open_windows:
                    debug("Closing display", disp.get_name())
                    disp.close()


    def get_timestamp(self):
        return self._timestamp

    def get_startup_notification_id(self):
        """
        Always returns a byte string
        """
        return _make_startup_notification_id(self.get_timestamp())

    def get_display(self):
        """return the display name to show new windows on

        Always returns a byte string
        """
        return self._screen.make_display_name()

    def get_screen(self):
        return self._screen

    def present_window(self, window):
        """
        Show and present @window on the current
        workspace, screen & display as appropriate.

        @window: A Gtk.Window
        """
        window.set_screen(self.get_screen())
        window.present_with_time(self.get_timestamp())

class _internal_data (object):
    seq = 0
    current_event_time = 0

    @classmethod
    def inc_seq(cls):
        cls.seq = cls.seq + 1

def _make_startup_notification_id(time):
    _internal_data.inc_seq()
    return "%s-%d-%s_TIME%d" % ("kupfer", os.getpid(), _internal_data.seq, time)

def current_event_time():
    return (Gtk.get_current_event_time() or
            keybindings.get_current_event_time() or
            _internal_data.current_event_time)

def _parse_notify_id(startup_notification_id):
    """
    Return timestamp or 0 from @startup_notification_id
    """
    time = 0
    if "_TIME" in startup_notification_id:
        _ign, bstime = startup_notification_id.split("_TIME", 1)
        try:
            time = abs(int(bstime))
        except ValueError:
            pass
    return time

@contextlib.contextmanager
def using_startup_notify_id(notify_id):
    """
    Pass in a DESKTOP_STARTUP_ID

    with using_startup_notify_id(...) as time:
        pass
    
    The yelt object is the parsed timestamp
    """
    timestamp = _parse_notify_id(notify_id)
    if timestamp:
        Gdk.notify_startup_complete_with_id(notify_id)
    try:
        pretty.print_debug(__name__, "Using startup id", repr(notify_id), timestamp)
        _internal_data.current_event_time = timestamp
        yield timestamp
    finally:
        _internal_data.current_event_time = Gdk.CURRENT_TIME






import gzip
import hashlib
import itertools
import pickle
import os
import threading
import time
import weakref

from kupfer import config, pretty, scheduler
from kupfer import conspickle
from kupfer.obj import base, sources
from kupfer.core import pluginload

class InternalError (Exception):
    pass

class PeriodicRescanner (pretty.OutputMixin):
    """
    Periodically rescan a @catalog of sources

    Do first rescan after @startup seconds, then
    followup with rescans in @period.

    Each campaign of rescans is separarated by @campaign
    seconds
    """
    def __init__(self, period=5, startup=10, campaign=3600):
        self.startup = startup
        self.period = period
        self.campaign=campaign
        self.timer = scheduler.Timer()
        # Source -> time mapping
        self.latest_rescan_time = weakref.WeakKeyDictionary()
        self._min_rescan_interval = campaign//4

    def set_catalog(self, catalog):
        self.catalog = catalog
        self.cur = iter(self.catalog)
        self.output_debug("Registering new campaign, in %d s" % self.startup)
        self.timer.set(self.startup, self._new_campaign)
    
    def _new_campaign(self):
        self.output_info("Starting new campaign, interval %d s" % self.period)
        self.cur = iter(self.catalog)
        self.timer.set(self.period, self._periodic_rescan_helper)

    def _periodic_rescan_helper(self):
        # Advance until we find a source that was not recently rescanned
        for next in self.cur:
            oldtime = self.latest_rescan_time.get(next, 0)
            if (time.time() - oldtime) > self._min_rescan_interval:
                self.timer.set(self.period, self._periodic_rescan_helper)
                self._start_source_rescan(next)
                return
        # No source to scan found
        self.output_info("Campaign finished, pausing %d s" % self.campaign)
        self.timer.set(self.campaign, self._new_campaign)

    def rescan_now(self, source, force_update=False):
        "Rescan @source immediately"
        if force_update:
            # if forced update, we know that it was brought up to date
            self.latest_rescan_time[source] = time.time()
        self.rescan_source(source, force_update=force_update)

    def _start_source_rescan(self, source):
        self.latest_rescan_time[source] = time.time()
        if not source.is_dynamic():
            thread = threading.Thread(target=self.rescan_source, args=(source,))
            thread.setDaemon(True)
            thread.start()

    def rescan_source(self, source, force_update=True):
        list(source.get_leaves(force_update=force_update))

class SourcePickler (pretty.OutputMixin):
    """
    Takes care of pickling and unpickling Kupfer Sources.
    """
    format_version = 5
    name_template = "k%s-v%d.pickle.gz"

    def __init__(self):
        self.open = lambda f,mode: gzip.open(f, mode, compresslevel=3)

    @classmethod
    def should_use_cache(cls):
        return config.has_capability("CACHE")

    @classmethod
    def should_use_cache_for_source(cls, source):
        return cls.should_use_cache() and source.source_use_cache

    def rm_old_cachefiles(self):
        """Checks if there are old cachefiles from last version,
        and deletes those
        """
        for dpath, dirs, files in os.walk(config.get_cache_home()):
            # Look for files matching beginning and end of
            # name_template, with the previous file version
            chead, ctail = self.name_template.split("%s")
            ctail = ctail % ((self.format_version -1),)
            obsolete_files = []
            for cfile in files:
                if cfile.startswith(chead) and cfile.endswith(ctail):
                    cfullpath = os.path.join(dpath, cfile)
                    obsolete_files.append(cfullpath)
        if obsolete_files:
            self.output_info("Removing obsolete cache files:", sep="\n",
                    *obsolete_files)
            for fpath in obsolete_files:
                # be overly careful
                assert fpath.startswith(config.get_cache_home())
                assert "kupfer" in fpath
                os.unlink(fpath)

    def get_filename(self, source):
        """Return cache filename for @source"""
        # make sure we take the source name into account
        # so that we get a "break" when locale changes
        source_id = "%s%s%s" % (repr(source), str(source), source.version)
        hash_str = hashlib.md5(source_id.encode("utf-8")).hexdigest()
        filename = self.name_template % (hash_str, self.format_version)
        return os.path.join(config.get_cache_home(), filename)

    def unpickle_source(self, source):
        if not self.should_use_cache_for_source(source):
            return None

        cached = self._unpickle_source(self.get_filename(source))
        if not cached:
            return None

        # check consistency
        if source == cached:
            return cached
        else:
            self.output_debug("Cached version mismatches", source)
        return None
    def _unpickle_source(self, pickle_file):
        try:
            pfile = self.open(pickle_file, "rb")
        except IOError as e:
            return None
        try:
            source = pickle.loads(pfile.read())
            assert isinstance(source, base.Source), "Stored object not a Source"
            sname = os.path.basename
            self.output_debug("Loading", source, "from", sname(pickle_file))
        except (pickle.PickleError, Exception) as e:
            source = None
            self.output_info("Error loading %s: %s" % (pickle_file, e))
        finally:
            pfile.close()
        return source

    def pickle_source(self, source):
        if not self.should_use_cache_for_source(source):
            return None
        return self._pickle_source(self.get_filename(source), source)
    def _pickle_source(self, pickle_file, source):
        """
        When writing to a file, use pickle.dumps()
        and then write the file in one go --
        if the file is a gzip file, pickler's thousands
        of small writes are very slow
        """
        output = self.open(pickle_file, "wb")
        sname = os.path.basename
        self.output_debug("Storing", source, "as", sname(pickle_file))
        output.write(pickle.dumps(source, pickle.HIGHEST_PROTOCOL))
        output.close()
        return True

class SourceDataPickler (pretty.OutputMixin):
    """ Takes care of pickling and unpickling Kupfer Sources' configuration
    or data.

    The SourceDataPickler requires a protocol of three methods:

    config_save_name()
      Return an ascii name to be used as a token/key for the configuration

    config_save()
      Return an object to be saved as configuration

    config_restore(obj)
      Receive the configuration object `obj' to load
    """
    format_version = 2
    name_template = "config-%s-v%d.pickle"

    def __init__(self):
        self.open = open

    @classmethod
    def get_filename(cls, source):
        """Return filename for @source"""
        name = source.config_save_name()
        filename = cls.name_template % (name, cls.format_version)
        return config.save_config_file(filename)

    @classmethod
    def source_has_config(self, source):
        return getattr(source, "config_save_name", None)

    def load_source(self, source):
        data = self._load_data(self.get_filename(source))
        if not data:
            return True
        source.config_restore(data)

    def _load_data(self, pickle_file):
        try:
            pfile = self.open(pickle_file, "rb")
        except IOError as e:
            return None
        try:
            data = conspickle.BasicUnpickler.loads(pfile.read())
            sname = os.path.basename(pickle_file)
            self.output_debug("Loaded configuration from", sname)
            # self.output_debug(data)
        except (pickle.PickleError, Exception) as e:
            data = None
            self.output_error("Loading %s: %s" % (pickle_file, e))
        finally:
            pfile.close()
        return data

    def save_source(self, source):
        return self._save_data(self.get_filename(source), source)
    def _save_data(self, pickle_file, source):
        sname = os.path.basename(pickle_file)
        obj = source.config_save()
        try:
            data = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)
        except pickle.PickleError:
            import traceback
            self.output_error("Unable to save configuration for", source)
            self.output_error("Saving configuration raised an exception:")
            traceback.print_exc()
            self.output_error("Please file a bug report")
            data = None
        if data:
            self.output_debug("Storing configuration for", source, "as", sname)
            ## Write to temporary and rename into place
            tmp_pickle_file = "%s.%s" % (pickle_file, os.getpid())
            output = self.open(tmp_pickle_file, "wb")
            output.write(data)
            output.close()
            os.rename(tmp_pickle_file, pickle_file)
        return True

class SourceController (pretty.OutputMixin):
    """Control sources; loading, pickling, rescanning

    Call .add() to add sources.
    Call .initialize() before use commences.
    """
    def __init__(self):
        self.rescanner = PeriodicRescanner(period=3)
        self.sources = set()
        self.toplevel_sources = set()
        self.text_sources = set()
        self.content_decorators = {}
        self.action_decorators = {}
        self.action_generators = []
        self.plugin_object_map = weakref.WeakKeyDictionary()
        self.loaded_successfully = False
        self.did_finalize_sources = False
        self._pre_root = None

    def add(self, plugin_id, srcs, toplevel=False, initialize=False):
        self._invalidate_root()
        sources = set(self._try_restore(srcs))
        sources.update(srcs)

        self.sources.update(sources)
        if toplevel:
            self.toplevel_sources.update(sources)
        if initialize:
            self._initialize_sources(sources)
            self._cache_sources(sources)
            self.rescanner.set_catalog(self.sources)
        if plugin_id:
            self._register_plugin_objects(plugin_id, *sources)

    def set_toplevel(self, src, toplevel):
        assert src in self.sources, "Source is not tracked in SourceController"
        self._invalidate_root()
        if toplevel:
            self.toplevel_sources.add(src)
        else:
            self.toplevel_sources.discard(src)

    def _register_plugin_objects(self, plugin_id, *objects):
        "Register a plugin id mapping for @objects"
        for obj in objects:
            self.plugin_object_map[obj] = plugin_id
            pretty.print_debug(__name__, "Add", repr(obj))

    def _remove(self, src):
        self._invalidate_root()
        self.toplevel_sources.discard(src)
        self.sources.discard(src)
        self.rescanner.set_catalog(self.sources)
        self._finalize_source(src)
        pretty.print_debug(__name__, "Remove", repr(src))

    def get_plugin_id_for_object(self, obj):
        id_ = self.plugin_object_map.get(obj)
        #self.output_debug("Object", repr(obj), "has id", id_, id(obj))
        return id_

    def remove_objects_for_plugin_id(self, plugin_id):
        """Remove all objects for @plugin_id

        Return True if the catalog configuration changed
        """
        removed_source = False
        self.output_debug("Removing objects for plugin:", plugin_id)

        # sources
        for src in self.sources:
            if self.get_plugin_id_for_object(src) == plugin_id:
                self._remove(src)
                removed_source = True

        # all other objects
        def remove_matching_objects(collection, plugin_id):
            for obj in list(collection):
                if self.get_plugin_id_for_object(obj) == plugin_id:
                    collection.remove(obj)
                    pretty.print_debug(__name__, "Remove", repr(obj))

        remove_matching_objects(self.text_sources, plugin_id)

        for typ in self.content_decorators:
            remove_matching_objects(self.content_decorators[typ], plugin_id)

        for typ in self.action_decorators:
            remove_matching_objects(self.action_decorators[typ], plugin_id)

        remove_matching_objects(self.action_generators, plugin_id)

        return removed_source

    def get_sources(self):
        return self.sources

    def add_text_sources(self, plugin_id, srcs):
        self.text_sources.update(srcs)
        self._register_plugin_objects(plugin_id, *srcs)

    def get_text_sources(self):
        return self.text_sources

    def add_content_decorators(self, plugin_id, decos):
        for typ in decos:
            self.content_decorators.setdefault(typ, set()).update(decos[typ])
            self._register_plugin_objects(plugin_id, *decos[typ])

    def add_action_decorators(self, plugin_id, decos):
        for typ in decos:
            self.action_decorators.setdefault(typ, set()).update(decos[typ])
            self._register_plugin_objects(plugin_id, *decos[typ])
        for typ in self.action_decorators:
            self._disambiguate_actions(self.action_decorators[typ])

    def add_action_generator(self, plugin_id, agenerator):
        self.action_generators.append(agenerator)
        self._register_plugin_objects(plugin_id, agenerator)

    def _disambiguate_actions(self, actions):
        """Rename actions by the same name (adding a suffix)"""
        # FIXME: Disambiguate by plugin name, not python module name
        names = {}
        renames = set()
        for action in actions:
            name = str(action)
            if name in names:
                renames.add(names[name])
                renames.add(action)
            else:
                names[name] = action
        for action in renames:
            self.output_debug("Disambiguate Action %s" % (action, ))
            plugin_suffix = " (%s)" % (type(action).__module__.split(".")[-1], )
            if not action.name.endswith(plugin_suffix):
                action.name += plugin_suffix

    def __contains__(self, src):
        return src in self.sources
    def __getitem__(self, src):
        if not src in self:
            raise KeyError
        for s in self.sources:
            if s == src:
                return s
    @property
    def root(self):
        """Get the root source of catalog"""
        if len(self.sources) == 1:
            root_catalog, = self.sources
        elif len(self.sources) > 1:
            firstlevel = self._firstlevel
            root_catalog = sources.MultiSource(firstlevel)
        else:
            root_catalog = None
        return root_catalog

    def _invalidate_root(self):
        "The source root needs to be recalculated"
        self._pre_root = None

    @property
    def _firstlevel(self):
        if self._pre_root:
            return self._pre_root
        sourceindex = set(self.sources)
        kupfer_sources = sources.SourcesSource(self.sources)
        sourceindex.add(kupfer_sources)
        # Make sure firstlevel is ordered
        # So that it keeps the ordering.. SourcesSource first
        firstlevel = []
        firstlevel.append(sources.SourcesSource(sourceindex))
        firstlevel.extend(set(self.toplevel_sources))
        self._pre_root = firstlevel
        return firstlevel

    @classmethod
    def good_source_for_types(cls, s, types):
        """return whether @s provides good leaves for @types
        """
        provides = list(s.provides())
        if not provides:
            return True
        for t in provides:
            if issubclass(t, types):
                return True

    def root_for_types(self, types, extra_sources=[]):
        """
        Get root for a flat catalog of all catalogs
        providing at least Leaves of @types

        types: Iterable of classes
        extra_sources: Sources to include

        Take all sources which:
            Provide a type T so that it is a subclass
            to one in the set of types we want
        """
        types = tuple(types)
        firstlevel = set(extra_sources)
        # include the Catalog index since we want to include
        # the top of the catalogs (like $HOME)
        catalog_index = (sources.SourcesSource(self.sources), )
        for s in itertools.chain(self.sources, catalog_index):
            if self.good_source_for_types(s, types):
                firstlevel.add(s)
        return sources.MultiSource(firstlevel)

    def get_canonical_source(self, source):
        "Return the canonical instance for @source"
        # check if we already have source, then return that
        if source in self:
            return self[source]
        else:
            source.initialize()
            return source

    def get_contents_for_leaf(self, leaf, types=None):
        """Iterator of content sources for @leaf,
        providing @types (or None for all)"""
        for typ in self.content_decorators:
            if not isinstance(leaf, typ):
                continue
            for content in self.content_decorators[typ]:
                dsrc = None
                with pluginload.exception_guard(
                        content,
                        self._remove_source,
                        content,
                        is_decorator=True):
                    dsrc = content.decorate_item(leaf)
                if dsrc:
                    if types and not self.good_source_for_types(dsrc, types):
                        continue
                    yield self.get_canonical_source(dsrc)

    def get_actions_for_leaf(self, leaf):
        for typ in self.action_decorators:
            if isinstance(leaf, typ):
                for act in self.action_decorators[typ]:
                    yield act
        for agenerator in self.action_generators:
            for action in agenerator.get_actions_for_leaf(leaf):
                yield action

    def decorate_object(self, obj, action=None):
        if hasattr(obj, "has_content") and not obj.has_content():
            types = tuple(action.object_types()) if action else ()
            contents = list(self.get_contents_for_leaf(obj, types))
            content = contents[0] if contents else None
            if len(contents) > 1:
                content = sources.SourcesSource(contents, name=str(obj),
                        use_reprs=False)
            obj.add_content(content)

    def finalize(self):
        "Finalize all sources, equivalent to deactivating all sources"
        for src in self.sources:
            src.finalize()
        self.did_finalize_sources = True

    def save_cache(self):
        "Save all caches (non-important data)"
        if not self.did_finalize_sources:
            raise InternalError("Called save_cache without finalize!")
        if self.loaded_successfully:
            self._pickle_sources(self.sources)
        else:
            self.output_debug("Not writing cache on failed load")

    def save_data(self):
        "Save (important) user data/configuration"
        if not self.loaded_successfully:
            self.output_info("Not writing configuration on failed load")
            return
        configsaver = SourceDataPickler()
        for source in self.sources:
            if configsaver.source_has_config(source):
                self._save_source(source, pickler=configsaver)

    @classmethod
    def _save_source(self, source, pickler=None):
        configsaver = pickler or SourceDataPickler()
        configsaver.save_source(source)

    def _finalize_source(self, source):
        "Either save config, or save cache for @source"
        source.finalize()
        if SourceDataPickler.source_has_config(source):
            self._save_source(source)
        elif not source.is_dynamic():
            self._pickle_source(source)

    def _pickle_sources(self, sources):
        sourcepickler = SourcePickler()
        sourcepickler.rm_old_cachefiles()
        for source in sources:
            if (source.is_dynamic() or
                SourceDataPickler.source_has_config(source)):
                continue
            self._pickle_source(source, pickler=sourcepickler)

    @classmethod
    def _pickle_source(self, source, pickler=None):
        sourcepickler = pickler or SourcePickler()
        sourcepickler.pickle_source(source)

    def _try_restore(self, sources):
        """
        Try to restore the source that is equivalent to the
        "dummy" instance @source, from cache, or from saved configuration.
        yield the instances that succeed.
        """
        sourcepickler = SourcePickler()
        configsaver = SourceDataPickler()
        for source in set(sources):
            if configsaver.source_has_config(source):
                configsaver.load_source(source)
            else:
                source = sourcepickler.unpickle_source(source)
            if source:
                yield source

    def _remove_source(self, source, is_decorator=False):
        "Oust @source from catalog if any exception is raised"
        if not is_decorator:
            self.sources.discard(source)
            self.toplevel_sources.discard(source)
            source_type = type(source)
        else:
            source_type = source
        for typ in self.content_decorators:
            self.content_decorators[typ].discard(source_type)

    def initialize(self):
        "Initialize all sources and cache toplevel sources"
        self._initialize_sources(self.sources)
        self.rescanner.set_catalog(self.sources)
        self._cache_sources(self.toplevel_sources)
        self.loaded_successfully = True

    def _initialize_sources(self, sources):
        for src in set(sources):
            with pluginload.exception_guard(src, self._remove_source, src):
                src.initialize()

    def _cache_sources(self, sources):
        # Make sure that the toplevel sources are chached
        # either newly rescanned or the cache is fully loaded
        for src in set(sources):
            with pluginload.exception_guard(src, self._remove_source, src):
                self.rescanner.rescan_now(src, force_update=False)


_source_controller = None
def GetSourceController():
    global _source_controller
    if _source_controller is None:
        _source_controller = SourceController()
    return _source_controller



"""
The main logic for executing constructed commands.

A command is normally a tuple of (object, action, indirect object).
Where, of course, the indirect object is often not needed (in this module we
then pass None in its stead).

This code was once a shining machine; While adding the "comma trick" and
support for "multiple dispatch" was easy in the rest of the program, it shed
its casualties here: While the main process is simple, we deal here with all
the exceptions that are, at the moment, tacked on.

The ActionExecutionContext (ACE) keeps track of its nested invocation, so that
we can catch the results of commands executed inside other commands. The
delegation mechanism allows a user of the ACE to indicate that the result of
the command should be passed on from the earlier (more nested) invocation.

Multiple dispatch is straightforward if the action implements the multiple
dispatch protocol. Is the protocol not implemented, the command is simply
"multiplied out": executed once for each object, or once for each combination
of object and indirect object.

With multiple command execution (and delegation), we must then process and
merge multiple return values.
"""


import collections
import contextlib
import itertools
import sys

from gi.repository import GObject

from kupfer import pretty
from kupfer import task
from kupfer import uiutils
from kupfer.objects import OperationError
from kupfer.obj.base import Leaf, Source
from kupfer.obj.objects import SourceLeaf
from kupfer.obj.sources import MultiSource
from kupfer.obj.compose import MultipleLeaf

RESULT_NONE, RESULT_OBJECT, RESULT_SOURCE, RESULT_ASYNC = (1, 2, 3, 4)
RESULTS_SYNC = (RESULT_OBJECT, RESULT_SOURCE)

_MAX_LAST_RESULTS = 10

_action_exec_context = None
def DefaultActionExecutionContext():
    global _action_exec_context
    if _action_exec_context is None:
        _action_exec_context = ActionExecutionContext()
    return _action_exec_context

class ActionExecutionError (Exception):
    pass

def _get_leaf_members(leaf):
    """
    Return an iterator to members of @leaf, if it is a multiple leaf
    """
    # NOTE : This function duplicates one in core/actionlogic.py
    try:
        return leaf.get_multiple_leaf_representation()
    except AttributeError:
        return (leaf, )

def _is_multiple(leaf):
    return hasattr(leaf, "get_multiple_leaf_representation")

def _wants_context(action):
    return action.wants_context()

def activate_action(context, obj, action, iobj):
    """ Activate @action in simplest manner """
    kwargs = {}
    if _wants_context(action):
        kwargs['ctx'] = context
    if not _is_multiple(obj) and not _is_multiple(iobj):
        return _activate_action_single(obj, action, iobj, kwargs)
    else:
        return _activate_action_multiple(obj, action, iobj, kwargs)

def _activate_action_single(obj, action, iobj, kwargs):
    if action.requires_object():
        ret = action.activate(obj, iobj, **kwargs)
    else:
        ret = action.activate(obj, **kwargs)
    return ret

def _activate_action_multiple(obj, action, iobj, kwargs):
    if not hasattr(action, "activate_multiple"):
        iobjs = (None, ) if iobj is None else _get_leaf_members(iobj)
        return _activate_action_multiple_multiplied(_get_leaf_members(obj),
                action, iobjs, kwargs)

    if action.requires_object():
        ret = action.activate_multiple(_get_leaf_members(obj),
                _get_leaf_members(iobj), **kwargs)
    else:
        ret = action.activate_multiple(_get_leaf_members(obj), **kwargs)
    return ret

def _activate_action_multiple_multiplied(objs, action, iobjs, kwargs):
    """
    Multiple dispatch by "mulitplied" invocation of the simple activation

    Return an iterable of the return values.
    """
    rets = []
    for L in objs:
        for I in iobjs:
            ret = _activate_action_single(L, action, I, kwargs)
            rets.append(ret)
    ctx = DefaultActionExecutionContext()
    ret = ctx._combine_action_result_multiple(action, rets)
    return ret

def parse_action_result(action, ret):
    """Return result type for @action and return value @ret"""
    def valid_result(ret):
        return ret and (not hasattr(ret, "is_valid") or ret.is_valid())

    # handle actions returning "new contexts"
    res = RESULT_NONE
    if action.is_factory() and valid_result(ret):
        res = RESULT_SOURCE
    if action.has_result() and valid_result(ret):
        res = RESULT_OBJECT
    elif action.is_async() and valid_result(ret):
        res = RESULT_ASYNC
    return res

def parse_late_action_result(action, ret):
    # Late result is assumed to be a Leaf (Object) result
    # by default for backward compat.
    #
    # It is also allowed to be a Source
    def valid_result(ret):
        return ret and (not hasattr(ret, "is_valid") or ret.is_valid())

    res = RESULT_NONE
    if isinstance(ret, Source) and valid_result(ret):
        res = RESULT_SOURCE
    elif valid_result(ret):
        res = RESULT_OBJECT
    return res

class ExecutionToken (object):
    """
    A token object that an ``Action`` carries with it
    from ``activate``.

    Must be used for access to current execution context,
    and to access the environment.
    """
    def __init__(self, aectx, async_token, ui_ctx):
        self._aectx = aectx
        self._token = async_token
        self._ui_ctx = ui_ctx

    def register_late_result(self, result_object, show=True):
        self._aectx.register_late_result(self._token, result_object, show=show,
                                         ctxenv=self._ui_ctx)

    def register_late_error(self, exc_info=None):
        self._aectx.register_late_error(self._token, exc_info)

    def delegated_run(self, *objs):
        return self._aectx.run(*objs, delegate=True, ui_ctx=self._ui_ctx)

    @property
    def environment(self):
        """This is a property for the current environment,
        acess env variables like this::

            ctx.environment.get_timestamp()

        Raises RuntimeError when not available.
        """
        if self._ui_ctx is not None:
            return self._ui_ctx
        else:
            raise RuntimeError("Environment Context not available")

class ActionExecutionContext (GObject.GObject, pretty.OutputMixin):
    """
    command-result (result_type, result)
        Emitted when a command is carried out, with its resulting value
    """
    __gtype_name__ = "ActionExecutionContext"
    def __init__(self):
        GObject.GObject.__init__(self)
        self.task_runner = task.TaskRunner(end_on_finish=False)
        self._nest_level = 0
        self._delegate = False
        self._command_counter = itertools.count()
        self.last_command_id = -1
        self.last_command = None
        self.last_executed_command = None
        self.last_results = collections.deque([], _MAX_LAST_RESULTS)

    def check_valid(self, obj, action, iobj):
        pass

    @contextlib.contextmanager
    def _nesting(self):
        try:
            self._nest_level += 1
            self._delegate = False
            yield
        finally:
            self._nest_level -= 1

    def _is_nested(self):
        return self._nest_level

    @contextlib.contextmanager
    def _error_conversion(self, *cmdtuple):
        try:
            yield
        except OperationError:
            self._do_error_conversion(cmdtuple, sys.exc_info())

    def _do_error_conversion(self, cmdtuple, exc_info):
        if not self.operation_error(exc_info, cmdtuple):
            raise
        etype, value, tb = exc_info
        raise ActionExecutionError(value).with_traceback(tb)

    def get_async_token(self):
        """Get an action execution for current execution

        Return a token for the currently active command execution.
        The token must be used for posting late results or late errors.
        """
        return (self.last_command_id, self.last_executed_command)

    def make_execution_token(self, ui_ctx):
        """
        Return an ExecutionToken for @self and @ui_ctx
        """
        return ExecutionToken(self, self.get_async_token(), ui_ctx)

    def operation_error(self, exc_info, cmdtuple):
        "Error when executing action. Return True when error was handled"
        if self._is_nested():
            return
        etype, value, tb = exc_info
        obj, action, iobj = cmdtuple
        # TRANS: When an error occurs in an action to be carried out,
        # TRANS: then this is the heading of the error notification
        return uiutils.show_notification(
                _("Could not to carry out '%s'") % action,
                str(value), icon_name="kupfer")

    def register_late_error(self, token, exc_info=None):
        "Register an error in exc_info. The error must be an OperationError"
        if exc_info is None:
            exc_info = sys.exc_info()
        if isinstance(exc_info, Exception):
            exc_info = (type(exc_info), exc_info, None)
        command_id, cmdtuple = token
        self._do_error_conversion(cmdtuple, exc_info)

    def register_late_result(self, token, result, show=True, ctxenv=None):
        """Register a late result

        Result must be a Leaf (as in result object, not factory or async)

        If @show, possibly display the result to the user.
        """
        self.output_debug("Late result", repr(result), "for", token)
        command_id, (_ign1, action, _ign2) = token
        if result is None:
            raise ActionExecutionError("Late result from %s was None" % action)
        assert isinstance(result, (Leaf, Source))
        res_name = str(result)
        res_desc = result.get_description()
        if res_desc:
            description = "%s (%s)" % (res_name, res_desc)
        else:
            description = res_name

        # If only registration was requsted, remove the command id info
        if not show:
            command_id = -1

        result_type = parse_late_action_result(action, result)

        self.output_debug("late-command-result", command_id, result_type, result, ctxenv)

        if result_type == RESULT_NONE:
            return
        uiutils.show_notification(_('"%s" produced a result') % action, description)

        self.emit("late-command-result", command_id, result_type, result, ctxenv)
        self._append_result(result_type, result)

    def _append_result(self, res_type, result):
        if res_type == RESULT_OBJECT:
            self.last_results.append(result)

    def run(self, obj, action, iobj, delegate=False, ui_ctx=None):
        """
        Activate the command (obj, action, iobj), where @iobj may be None

        Return a tuple (DESCRIPTION; RESULT)

        If a command carries out another command as part of its execution,
        and wishes to delegate to it, pass True for @delegate.
        """
        self.last_command_id = next(self._command_counter)
        self.last_executed_command = (obj, action, iobj)

        if not action or not obj:
            raise ActionExecutionError("Primary Object and Action required")
        if iobj is None and action.requires_object():
            raise ActionExecutionError("%s requires indirect object" % action)
        self.output_debug(repr(obj), repr(action), repr(iobj), repr(ui_ctx))

        # The execution token object for the current invocation
        execution_token = self.make_execution_token(ui_ctx)
        with self._error_conversion(obj, action, iobj):
            with self._nesting():
                ret = activate_action(execution_token, obj, action, iobj)

        # remember last command, but not delegated commands.
        if not delegate:
            self.last_command = self.last_executed_command

        # Delegated command execution was previously requested: we take
        # the result of the nested execution context
        if self._delegate:
            res, ret = ret
            return self._return_result(res, ret, ui_ctx)

        res = parse_action_result(action, ret)
        if res == RESULT_ASYNC:
            # Register the task then "clear" the result
            self.output_debug("Registering async task", ret)
            self.task_runner.add_task(ret)
            res, ret = RESULT_NONE, None

        # Delegated command execution was requested: we pass
        # through the result of the action to the parent execution context
        if delegate and self._is_nested():
            self._delegate = True

        return self._return_result(res, ret, ui_ctx)

    def _return_result(self, res, ret, ui_ctx):
        if not self._is_nested():
            self._append_result(res, ret)
            self.emit("command-result", res, ret, ui_ctx)
        return res, ret


    def _combine_action_result_multiple(self, action, retvals):
        self.output_debug("Combining", repr(action), retvals,
                "delegate=%s" % self._delegate)

        def _make_retvalue(res, values):
            "Construct a return value for type res"
            if res == RESULT_SOURCE:
                return values[0] if len(values) == 1 else MultiSource(values)
            if res == RESULT_OBJECT:
                return values[0] if len(values) == 1 else MultipleLeaf(values)
            if res == RESULT_ASYNC:
                # Register all tasks now, and return None upwards
                for task_ in values:
                    self.output_debug("Registering async task", task_)
                    self.task_runner.add_task(task_)
            return None

        if not self._delegate:
            values = []
            res = RESULT_NONE
            for ret in retvals:
                res_type = parse_action_result(action, ret)
                if res_type != RESULT_NONE:
                    values.append(ret)
                    res = res_type
            return _make_retvalue(res, values)
        else:
            # Re-parse result values
            res = RESULT_NONE
            resmap = {}
            for ret in retvals:
                if ret is None:
                    continue
                res_type, ret_obj = ret
                if res_type != RESULT_NONE:
                    res = res_type
                    resmap.setdefault(res_type, []).append(ret_obj)

            # register tasks
            tasks = resmap.pop(RESULT_ASYNC, [])
            _make_retvalue(RESULT_ASYNC, tasks)

            if len(resmap) == 1:
                # Return the only of the Source or Object case
                key, values = list(resmap.items())[0]
                return key, _make_retvalue(key, values)
            elif len(resmap) > 1:
                # Put the source in a leaf and return a multiple leaf
                source = _make_retvalue(RESULT_SOURCE, resmap[RESULT_SOURCE])
                objects = resmap[RESULT_OBJECT]
                objects.append(SourceLeaf(source))
                return RESULT_OBJECT, _make_retvalue(RESULT_OBJECT, objects)
            return RESULT_NONE, None


# Signature: Action result type, action result, gui_context
GObject.signal_new("command-result", ActionExecutionContext,
        GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_INT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT))

# Signature: Command ID, Action result type, action result, gui_context
GObject.signal_new("late-command-result", ActionExecutionContext,
        GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_INT, GObject.TYPE_INT,
            GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT))



"""
Stub implementation of HTTPS connections.
"""

class VerifiedHTTPSConnection (object):
    "implementation stub"
    def __init__(self, host, *args, **kwargs):
        pass
    @classmethod
    def is_ssl_supported(cls):
        return False

def is_supported():
    return VerifiedHTTPSConnection and VerifiedHTTPSConnection.is_ssl_supported()



"""
Persistent Globally Unique Indentifiers for KupferObjects.

Some objects are assigned identifiers by reference, some are assigned
identifiers containing the whole object data (SerializedObject).

SerializedObject is a saved representation of a KupferObject, i.e. a
data model user-level object.

We unpickle SerializedObjects in an especially conservative way: new
module loading is always refused; this way, we avoid loading parts of
the program that we didn't wish to activate.
"""

import contextlib
import pickle

from kupfer import pretty
from kupfer.core import actioncompat
from kupfer.core import qfurl
from kupfer.core.sources import GetSourceController
from kupfer.conspickle import ConservativeUnpickler

__all__ = [
    "SerializedObject", "SERIALIZABLE_ATTRIBUTE",
    "resolve_unique_id", "resolve_action_id", "get_unique_id", "is_reference",
]


SERIALIZABLE_ATTRIBUTE = "serializable"


class SerializedObject (object):
    # treat the serializable attribute as a version number, defined on the class
    def __init__(self, obj):
        self.version = getattr(obj, SERIALIZABLE_ATTRIBUTE)
        self.data = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)
    def __hash__(self):
        return hash(self.data)

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.data == other.data and
                self.version == other.version)
    def reconstruct(self):
        obj = ConservativeUnpickler.loads(self.data)
        if self.version != getattr(obj, SERIALIZABLE_ATTRIBUTE):
            raise ValueError("Version mismatch for reconstructed %s" % obj)
        return obj

def get_unique_id(obj):
    if obj is None:
        return None
    if hasattr(obj, "qf_id"):
        return str(qfurl.qfurl(obj))
    if getattr(obj, SERIALIZABLE_ATTRIBUTE, None) is not None:
        try:
            return SerializedObject(obj)
        except pickle.PicklingError as exc:
            pretty.print_error(__name__, type(exc).__name__, exc)
            return None
    return repr(obj)

def is_reference(puid):
    "Return True if @puid is a reference-type ID"
    return not isinstance(puid, SerializedObject)

# A Context manager to block recursion when seeking inside a
# catalog; we have a stack (@_excluding) of the sources we
# are visiting, and nested context with the _exclusion
# context manager

_excluding = []
@contextlib.contextmanager
def _exclusion(src):
    try:
        _excluding.append(src)
        yield
    finally:
        _excluding.pop()

def _is_currently_excluding(src):
    return src is not None and src in _excluding

def _find_obj_in_catalog(puid, catalog):
    if puid.startswith(qfurl.QFURL_SCHEME):
        qfu = qfurl.qfurl(url=puid)
        return qfu.resolve_in_catalog(catalog)
    for src in catalog:
        if _is_currently_excluding(src):
            continue
        with _exclusion(src):
            for obj in src.get_leaves():
                if repr(obj) == puid:
                    return obj
    return None

def resolve_unique_id(puid, excluding=None):
    """
    Resolve unique id @puid

    The caller (if a Source) should pass itself as @excluding,
    so that recursion into itself is avoided.
    """
    if excluding is not None:
        with _exclusion(excluding):
            return resolve_unique_id(puid, None)

    if puid is None:
        return None
    if isinstance(puid, SerializedObject):
        try:
            return puid.reconstruct()
        except Exception as exc:
            pretty.print_debug(__name__, type(exc).__name__, exc)
            return None
    sc = GetSourceController()
    obj = _find_obj_in_catalog(puid, sc._firstlevel)
    if obj is not None:
        return obj
    other_sources = set(sc.sources) - set(sc._firstlevel)
    obj = _find_obj_in_catalog(puid, other_sources)
    return obj

def resolve_action_id(puid, for_item=None):
    if puid is None:
        return None
    if isinstance(puid, SerializedObject):
        return resolve_unique_id(puid)
    get_action_id = repr
    sc = GetSourceController()
    if for_item is not None:
        for action in actioncompat.actions_for_item(for_item, sc):
            if get_unique_id(action) == puid:
                return action
    for item_type, actions in sc.action_decorators.items():
        for action in actions:
            if get_action_id(action) == puid:
                return action
    pretty.print_debug(__name__, "Unable to resolve %s (%s)" % (puid, for_item))
    return None



# -*- coding: UTF-8 -*-

"""
Copyright 2007--2009 Ulrik Sverdrup <ulrik.sverdrup@gmail.com>

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

import os
from os import path
import zlib

from gi.repository import GLib, Gio

from kupfer import icons, launch, utils
from kupfer import pretty
from kupfer.obj.base import Leaf, Action
from kupfer.obj.base import InvalidDataError, OperationError
from kupfer.obj import fileactions
from kupfer.interface import TextRepresentation
from kupfer.kupferstring import tounicode
from kupfer.version import DESKTOP_ID

def ConstructFileLeafTypes():
    """ Return a seq of the Leaf types returned by ConstructFileLeaf"""
    yield FileLeaf
    yield AppLeaf

def ConstructFileLeaf(obj):
    """
    If the path in @obj points to a Desktop Item file,
    return an AppLeaf, otherwise return a FileLeaf
    """
    root, ext = path.splitext(obj)
    if ext == ".desktop":
        try:
            return AppLeaf(init_path=obj)
        except InvalidDataError:
            pass
    return FileLeaf(obj)

def _directory_content(dirpath, show_hidden):
    from kupfer.obj.sources import DirectorySource
    return DirectorySource(dirpath, show_hidden)

def _as_gfile(file_path):
    return Gio.File.new_for_path(file_path)

def _display_name(g_file):
    info = \
    g_file.query_info(Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
                      Gio.FileQueryInfoFlags.NONE,
                      None)
    return info.get_attribute_string(Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME)

class FileLeaf (Leaf, TextRepresentation):
    """
    Represents one file: the represented object is a bytestring (important!)
    """
    serializable = 1

    def __init__(self, obj, name=None, alias=None):
        """Construct a FileLeaf

        The display name of the file is normally derived from the full path,
        and @name should normally be left unspecified.

        @obj: byte string (file system encoding)
        @name: unicode name or None for using basename
        """
        if obj is None:
            raise InvalidDataError("File path for %s may not be None" % name)
        # Use glib filename reading to make display name out of filenames
        # this function returns a `unicode` object
        if not name:
            unicode_path = tounicode(obj)
            name = GLib.filename_display_basename(unicode_path)
        super(FileLeaf, self).__init__(obj, name)
        if alias:
            self.kupfer_add_alias(alias)

    @classmethod
    def from_uri(cls, uri):
        """
        Construct a FileLeaf

        uri: A local uri

        Return FileLeaf if it is supported, else None
        """
        gfile = Gio.File.new_for_uri(uri)
        p = gfile.get_path()
        if p:
            return cls(p)
        else:
            return None

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        try:
            return (type(self) == type(other) and
                    str(self) == str(other) and
                    path.samefile(self.object, other.object))
        except OSError as exc:
            pretty.print_debug(__name__, exc)
            return False

    def repr_key(self):
        return self.object

    def canonical_path(self):
        """Return the true path of the File (without symlinks)"""
        return path.realpath(self.object)

    def is_valid(self):
        return os.access(self.object, os.R_OK)

    def _is_executable(self):
        return os.access(self.object, os.R_OK | os.X_OK)

    def is_dir(self):
        return path.isdir(self.object)

    def get_text_representation(self):
        return GLib.filename_display_name(self.object)

    def get_urilist_representation(self):
        return [self.get_gfile().get_uri()]

    def get_gfile(self):
        """
        Return a Gio.File of self
        """
        return _as_gfile(self.object)

    def get_description(self):
        return utils.get_display_path_for_bytestring(self.canonical_path())

    def get_actions(self):
        return fileactions.get_actions_for_file(self)

    def has_content(self):
        return self.is_dir() or Leaf.has_content(self)
    def content_source(self, alternate=False):
        if self.is_dir():
            return _directory_content(self.object, alternate)
        else:
            return Leaf.content_source(self)

    def get_thumbnail(self, width, height):
        if self.is_dir(): return None
        return icons.get_thumbnail_for_gfile(self.get_gfile(), width, height)

    def get_gicon(self):
        return icons.get_gicon_for_file(self.object)
    def get_icon_name(self):
        if self.is_dir():
            return "folder"
        else:
            return "text-x-generic"

    def get_content_type(self):
        ret, uncertain = Gio.content_type_guess(self.object, None)
        if not uncertain:
            return ret
        content_attr = Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE
        gfile = self.get_gfile()
        if not gfile.query_exists(None):
            return None
        info = gfile.query_info(content_attr, Gio.FileQueryInfoFlags.NONE, None)
        content_type = info.get_attribute_string(content_attr)
        return content_type

    def is_content_type(self, ctype):
        """
        Return True if this file is of the type ctype

        ctype: A mime type, can have wildcards like 'image/*'
        """
        predicate = Gio.content_type_is_a
        ctype_guess, uncertain = Gio.content_type_guess(self.object, None)
        ret = predicate(ctype_guess, ctype)
        if ret or not uncertain:
            return ret
        content_attr = Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE
        gfile = self.get_gfile()
        if not gfile.query_exists(None):
            return
        info = gfile.query_info(content_attr, Gio.FileQueryInfoFlags.NONE, None)
        content_type = info.get_attribute_string(content_attr)
        return predicate(content_type, ctype)


class SourceLeaf (Leaf):
    def __init__(self, obj, name=None):
        """Create SourceLeaf for source @obj"""
        if not name:
            name = str(obj)
        Leaf.__init__(self, obj, name)
    def has_content(self):
        return True

    def repr_key(self):
        return repr(self.object)

    def content_source(self, alternate=False):
        return self.object

    def get_description(self):
        return self.object.get_description()

    @property
    def fallback_icon_name(self):
        return self.object.fallback_icon_name

    def get_gicon(self):
        return self.object.get_gicon()

    def get_icon_name(self):
        return self.object.get_icon_name()

class AppLeaf (Leaf):
    def __init__(self, item=None, init_path=None, app_id=None, require_x=True):
        """Try constructing an Application for GAppInfo @item,
        for file @path or for package name @app_id.

        @require_x: require executable file
        """
        self.init_item = item
        self.init_path = init_path
        self.init_item_id = app_id and app_id + ".desktop"
        # finish will raise InvalidDataError on invalid item
        self.finish(require_x)
        Leaf.__init__(self, self.object, self.object.get_name())
        self._add_aliases()

    def _add_aliases(self):
        # find suitable alias
        # use package name: non-extension part of ID
        lowername = str(self).lower()
        package_name = self._get_package_name()
        if package_name and package_name not in lowername:
            self.kupfer_add_alias(package_name)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                self.get_id() == other.get_id())

    def __getstate__(self):
        self.init_item_id = self.object and self.object.get_id()
        state = dict(vars(self))
        state["object"] = None
        state["init_item"] = None
        return state

    def __setstate__(self, state):
        vars(self).update(state)
        self.finish()

    def finish(self, require_x=False):
        """Try to set self.object from init's parameters"""
        item = None
        if self.init_item:
            item = self.init_item
        else:
            # Construct an AppInfo item from either path or item_id
            try:
                if self.init_path and (
                   not require_x or os.access(self.init_path, os.X_OK)):
                    # serilizable if created from a "loose file"
                    self.serializable = 1
                    item = Gio.DesktopAppInfo.new_from_filename(self.init_path)
                elif self.init_item_id:
                        item = Gio.DesktopAppInfo.new(self.init_item_id)
            except TypeError:
                pretty.print_debug(__name__, "Application not found:",
                        self.init_item_id, self.init_path)
                raise InvalidDataError
        self.object = item
        if not self.object:
            raise InvalidDataError

    def repr_key(self):
        return self.get_id()

    def _get_package_name(self):
        return GLib.filename_display_basename(self.get_id())

    def launch(self, files=(), paths=(), activate=False, ctx=None):
        """
        Launch the represented applications

        @files: a seq of GFiles (Gio.File)
        @paths: a seq of bytestring paths
        @activate: activate instead of start new
        """
        try:
            return launch.launch_application(self.object, files=files,
                                             paths=paths, activate=activate,
                                             desktop_file=self.init_path,
                                             screen=ctx and ctx.environment.get_screen())
        except launch.SpawnError as exc:
            raise OperationError(exc)

    def get_id(self):
        """Return the unique ID for this app.

        This is the GIO id "gedit.desktop" minus the .desktop part for
        system-installed applications.
        """
        return launch.application_id(self.object, self.init_path)

    def get_actions(self):
        id_ = self.get_id()
        if id_ == DESKTOP_ID:
            return
        if launch.application_is_running(id_):
            yield Launch(_("Go To"), is_running=True)
            yield CloseAll()
        else:
            yield Launch()
        yield LaunchAgain()

    def get_description(self):
        # Use Application's description, else use executable
        # for "file-based" applications we show the path
        app_desc = tounicode(self.object.get_description())
        ret = tounicode(app_desc if app_desc else self.object.get_executable())
        if self.init_path:
            app_path = utils.get_display_path_for_bytestring(self.init_path)
            return "(%s) %s" % (app_path, ret)
        return ret

    def get_gicon(self):
        return self.object.get_icon()

    def get_icon_name(self):
        return "exec"

class OpenUrl (Action):
    action_accelerator = "o"
    rank_adjust = 5
    def __init__(self, name=None):
        if not name:
            name = _("Open URL")
        super(OpenUrl, self).__init__(name)

    def activate(self, leaf):
        url = leaf.object
        self.open_url(url)

    def open_url(self, url):
        utils.show_url(url)

    def get_description(self):
        return _("Open URL with default viewer")

    def get_icon_name(self):
        return "forward"

class Launch (Action):
    """ Launches an application (AppLeaf) """
    action_accelerator = "o"
    rank_adjust = 5
    def __init__(self, name=None, is_running=False, open_new=False):
        """
        If @is_running, style as if the app is running (Show application)
        If @open_new, always start a new instance.
        """
        if not name:
            name = _("Launch")
        Action.__init__(self, name)
        self.is_running = is_running
        self.open_new = open_new

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        leaf.launch(activate=not self.open_new, ctx=ctx)

    def get_description(self):
        if self.is_running:
            return _("Show application window")
        return _("Launch application")

    def get_icon_name(self):
        if self.is_running:
            return "go-jump"
        return "kupfer-launch"

class LaunchAgain (Launch):
    action_accelerator = None
    rank_adjust = 0
    def __init__(self, name=None):
        if not name:
            name = _("Launch Again")
        Launch.__init__(self, name, open_new=True)
    def item_types(self):
        yield AppLeaf
    def valid_for_item(self, leaf):
        return launch.application_is_running(leaf.get_id())
    def get_description(self):
        return _("Launch another instance of this application")

class CloseAll (Action):
    """Attempt to close all application windows"""
    rank_adjust = -10
    def __init__(self):
        Action.__init__(self, _("Close"))
    def activate(self, leaf):
        return launch.application_close_all(leaf.get_id())
    def item_types(self):
        yield AppLeaf
    def valid_for_item(self, leaf):
        return launch.application_is_running(leaf.get_id())
    def get_description(self):
        return _("Attempt to close all application windows")
    def get_icon_name(self):
        return "window-close"

class UrlLeaf (Leaf, TextRepresentation):
    serializable = 1
    def __init__(self, obj, name):
        super(UrlLeaf, self).__init__(obj, name or obj)
        if obj != name:
            self.kupfer_add_alias(obj)

    def get_actions(self):
        yield OpenUrl()

    def get_description(self):
        return self.object

    def get_icon_name(self):
        return "text-html"

class RunnableLeaf (Leaf):
    """Leaf where the Leaf is basically the action itself,
    for items such as Quit, Log out etc.
    """
    def __init__(self, obj=None, name=None):
        Leaf.__init__(self, obj, name)
    def get_actions(self):
        yield Perform()
    def run(self, ctx=None):
        raise NotImplementedError
    def wants_context(self):
        """ Return ``True`` if you want the actions' execution
        context passed as ctx= in RunnableLeaf.run
        """
        return False
    def repr_key(self):
        return ""
    def get_gicon(self):
        iname = self.get_icon_name()
        if iname:
            return icons.get_gicon_with_fallbacks(None, (iname, ))
        return icons.ComposedIcon("kupfer-object", "kupfer-execute")
    def get_icon_name(self):
        return ""

class Perform (Action):
    """Perform the action in a RunnableLeaf"""
    action_accelerator = "o"
    rank_adjust = 5
    def __init__(self, name=None):
        # TRANS: 'Run' as in Perform a (saved) command
        if not name: name = _("Run")
        super(Perform, self).__init__(name=name)
    def wants_context(self):
        return True
    def activate(self, leaf, ctx):
        if leaf.wants_context():
            return leaf.run(ctx)
        else:
            return leaf.run()
    def get_description(self):
        return _("Perform command")

class TextLeaf (Leaf, TextRepresentation):
    """Represent a text query
    The represented object is a unicode string
    """
    serializable = 1
    def __init__(self, text, name=None):
        """@text *must* be unicode or UTF-8 str"""
        text = tounicode(text)
        if not name:
            name = self.get_first_text_line(text)
        if len(text) == 0 or not name:
            name = _("(Empty Text)")
        Leaf.__init__(self, text, name)

    def get_actions(self):
        return ()

    def repr_key(self):
        return zlib.crc32(self.object.encode("utf-8", "surrogateescape"))

    @classmethod
    def get_first_text_line(cls, text):
        firstline = None
        firstnl = text.find("\n")
        if firstnl != -1:
            firstline = text[:firstnl].strip()
            if not firstline:
                splut = text.split(None, 1)
                firstline = splut[0] if splut else text
        else:
            firstline = text
        if not firstline:
            firstline = text.strip("\n")
        return firstline

    def get_description(self):
        numlines = self.object.count("\n") + 1
        desc = self.get_first_text_line(self.object)

        # TRANS: This is description for a TextLeaf, a free-text search
        # TRANS: The plural parameter is the number of lines %(num)d
        return ngettext('"%(text)s"', '(%(num)d lines) "%(text)s"',
            numlines) % {"num": numlines, "text": desc }

    def get_icon_name(self):
        return "edit-select-all"




# TRANS: "Glob" is the matching files like a shell with "*.py" etc.
__kupfer_name__ = _("Glob")
__kupfer_actions__ = ("Glob",)
__description__ = _("Select objects using '*' and '?' as wildcards.")
__version__ = ""
__author__ = ""

import fnmatch
import re

from kupfer.objects import Action, TextLeaf, TextSource, Leaf, OperationError
from kupfer.obj.compose import MultipleLeaf

class Glob (Action):
    def __init__(self):
        Action.__init__(self, _("Glob"))

    def activate(self, obj, iobj):
        return self.activate_multiple((obj,), (iobj, ))

    def activate_multiple(self, objects, iobjects):
        ## Do case-insentive matching
        ## As a special case, understand '**/' prefix as recurive

        def get_subcatalog_matches(subcatalog, pat, recursive, paths):
            if len(paths) > 1000:
                raise OperationError("Globbing wayy too many objects")
            for content in subcatalog.content_source().get_leaves():
                if recursive and content.has_content():
                    get_subcatalog_matches(content, pat, recursive, paths)
                else:
                    if re.match(pat, str(content), flags=re.I):
                        paths.append(content)
        paths = []
        for iobj in iobjects:
            glob = iobj.object
            if glob.startswith('**/'):
                glob = glob[3:]
                recursive = True
            else:
                recursive = False
            pat = fnmatch.translate(glob)
            for obj in objects:
                get_subcatalog_matches(obj, pat, recursive, paths)
        if paths:
            return MultipleLeaf(paths)

    def has_result(self):
        return True
    def item_types(self):
        yield Leaf
    def valid_for_item(self, item):
        return item.has_content()
    def requires_object(self):
        return True
    def object_types(self):
        yield TextLeaf
    def object_source(self, for_item=None):
        return TextSource()
    def valid_object(self, iobj, for_item):
        return ('*' in iobj.object) or ('?' in iobj.object)

    def get_description(self):
        return __description__



from kupfer.obj.base import InvalidDataError, Source
from kupfer.obj.helplib import PicklingHelperMixin, FilesystemWatchMixin
from kupfer.obj.objects import AppLeaf

class AppLeafContentMixin (object):
    """
    Mixin for Source that correspond one-to-one with a AppLeaf

    This Mixin sees to that the Source is set as content for the application
    with id 'cls.appleaf_content_id', which may also be a sequence of ids.

    Source has to define the attribute appleaf_content_id and must
    inherit this mixin BEFORE the Source

    This Mixin defines:
    get_leaf_repr
    decorates_type,
    decorates_item
    """
    @classmethod
    def get_leaf_repr(cls):
        if not hasattr(cls, "_cached_leaf_repr"):
            cls._cached_leaf_repr = cls.__get_leaf_repr()
        return cls._cached_leaf_repr
    @classmethod
    def __get_appleaf_id_iter(cls):
        if isinstance(cls.appleaf_content_id, str):
            ids = (cls.appleaf_content_id, )
        else:
            ids = list(cls.appleaf_content_id)
        return ids
    @classmethod
    def __get_leaf_repr(cls):
        for appleaf_id in cls.__get_appleaf_id_iter():
            try:
                return AppLeaf(app_id=appleaf_id)
            except InvalidDataError:
                pass
    @classmethod
    def decorates_type(cls):
        return AppLeaf
    @classmethod
    def decorate_item(cls, leaf):
        if leaf == cls.get_leaf_repr():
            return cls()

class ApplicationSource(AppLeafContentMixin, Source, PicklingHelperMixin,
        FilesystemWatchMixin):
    pass




"""
This is a DuckDuckGo search plugin based on the Wikipedia search plugin
"""

__kupfer_name__ = _("DuckDuckGo Search")
__kupfer_sources__ = ()
__kupfer_actions__ = ("DuckDuckGoSearch",)
__description__ = _("Search the web securely with DuckDuckGo")
__version__ = "1.0"
__author__ = "Isaac Aggrey <isaac.aggrey@gmail.com>"

import urllib.parse

from kupfer.objects import Action, TextLeaf
from kupfer import utils

class DuckDuckGoSearch (Action):
    def __init__(self):
        Action.__init__(self, _("DuckDuckGo Search"))

    def activate(self, leaf):
        search_url = "https://duckduckgo.com/"
        query_url = search_url + "?" + urllib.parse.urlencode({"q" : leaf.object})
        utils.show_url(query_url)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search the web securely with DuckDuckGo")

    def get_icon_name(self):
        return "edit-find"



from time import time
import os
import pickle as pickle

from gi.repository import GLib, Gio

from kupfer import pretty, config
from kupfer import scheduler
from kupfer import desktop_launch
from kupfer.ui import uievents
from kupfer import terminal

from kupfer.desktop_launch import SpawnError

## NOTE: SpawnError  *should* be imported from this module

try:
    import gi
    gi.require_version("Wnck", "3.0")
    from gi.repository import Wnck
    Wnck.set_client_type(Wnck.ClientType.PAGER)
    if 'WAYLAND_DISPLAY' in os.environ:
        Wnck.Screen.get_default = lambda *x : None
except ValueError as e:
    pretty.print_info(__name__, "Disabling window tracking:", e)
    Wnck = None


default_associations = {
    "evince" : "Document Viewer",
    "file-roller" : "File Roller",
    #"gedit" : "Text Editor",
    "gnome-keyring-manager" : "Keyring Manager",
    "nautilus-browser" : "File Manager",
    "rhythmbox" : "Rhythmbox Music Player",
}


def application_id(app_info, desktop_file=None):
    """Return an application id (string) for GAppInfo @app_info"""
    app_id = app_info.get_id()
    if not app_id:
        app_id = desktop_file or ""
    if app_id.endswith(".desktop"):
        app_id = app_id[:-len(".desktop")]
    return app_id

def launch_application(app_info, files=(), uris=(), paths=(), track=True,
                       activate=True, desktop_file=None, screen=None):
    """
    Launch @app_rec correctly, using a startup notification

    you may pass in either a list of Gio.Files in @files, or 
    a list of @uris or @paths

    if @track, it is a user-level application
    if @activate, activate rather than start a new version

    @app_rec is either an GAppInfo or (GAppInfo, desktop_file_path) tuple

    Raises SpawnError on failed program start.
    """
    assert app_info

    if paths:
        files = [Gio.File.new_for_path(p) for p in paths]
    if uris:
        files = [Gio.File.new_for_uri(p) for p in uris]

    svc = GetApplicationsMatcherService()
    app_id = application_id(app_info, desktop_file)

    if activate and svc.application_is_running(app_id):
        svc.application_to_front(app_id)
        return True

    # An launch callback closure for the @app_id
    def application_launch_callback(argv, pid, notify_id, files, timestamp):
        is_terminal = terminal.is_known_terminal_executable(argv[0])
        if not is_terminal:
            svc.launched_application(app_id, pid)

    if track:
        launch_callback = application_launch_callback
    else:
        launch_callback = None

    try:
        desktop_launch.launch_app_info(app_info, files,
               timestamp=uievents.current_event_time(),
               desktop_file=desktop_file,
               launch_cb=launch_callback,
               screen=screen)
    except SpawnError:
        raise
    return True

def application_is_running(app_id):
    svc = GetApplicationsMatcherService()
    return svc.application_is_running(app_id)

def application_close_all(app_id):
    svc = GetApplicationsMatcherService()
    return svc.application_close_all(app_id)

class ApplicationsMatcherService (pretty.OutputMixin):
    """Handle launching applications and see if they still run.
    This is a learning service, since we have no first-class application
    object on the Linux desktop
    """
    def __init__(self):
        self.register = {}
        self._get_wnck_screen_windows_stacked()
        scheduler.GetScheduler().connect("finish", self._finish)
        self._load()

    @classmethod
    def _get_wnck_screen_windows_stacked(cls):
        if not Wnck:
            return ()
        screen = Wnck.Screen.get_default()
        if screen is None:
            return ()
        return screen.get_windows_stacked()

    def _get_filename(self):
        # Version 1: Up to incl v203
        # Version 2: Do not track terminals
        version = 2
        return os.path.join(config.get_cache_home(),
                "application_identification_v%d.pickle" % version)
    def _load(self):
        reg = self._unpickle_register(self._get_filename())
        self.register = reg if reg else default_associations
        # pretty-print register to debug
        if self.register:
            self.output_debug("Learned the following applications")
            self.output_debug("\n{\n%s\n}" % "\n".join(
                ("  %-30s : %s" % (k,v)
                    for k,v in self.register.items())
                ))
    def _finish(self, sched):
        self._pickle_register(self.register, self._get_filename())
    def _unpickle_register(self, pickle_file):
        try:
            pfile = open(pickle_file, "rb")
        except IOError as e:
            return None
        try:
            source = pickle.loads(pfile.read())
            assert isinstance(source, dict), "Stored object not a dict"
            self.output_debug("Reading from %s" % (pickle_file, ))
        except (pickle.PickleError, Exception) as e:
            source = None
            self.output_info("Error loading %s: %s" % (pickle_file, e))
        finally:
            pfile.close()
        return source

    def _pickle_register(self, reg, pickle_file):
        output = open(pickle_file, "wb")
        self.output_debug("Saving to %s" % (pickle_file, ))
        output.write(pickle.dumps(reg, pickle.HIGHEST_PROTOCOL))
        output.close()
        return True

    def _store(self, app_id, window):
        # FIXME: Store the 'res_class' name?
        application = window.get_application()
        res_class = window.get_class_group().get_res_class()
        store_name = application.get_name()
        self.register[app_id] = res_class
        self.output_debug("storing application", app_id, "as", store_name,
                          "res_class", res_class)

    def _has_match(self, app_id):
        return app_id in self.register

    def _is_match(self, app_id, window):
        application = window.get_application()
        res_class = window.get_class_group().get_res_class()
        reg_name = self.register.get(app_id)
        if reg_name and reg_name in (application.get_name(), res_class):
            return True
        if app_id in (application.get_name().lower(), res_class.lower()):
            return True
        return False

    def launched_application(self, app_id, pid):
        if self._has_match(app_id):
            return
        timeout = time() + 15
        GLib.timeout_add_seconds(2, self._find_application, app_id, pid, timeout)
        # and once later
        GLib.timeout_add_seconds(30, self._find_application, app_id, pid, timeout)

    def _find_application(self, app_id, pid, timeout):
        if self._has_match(app_id):
            return False
        #self.output_debug("Looking for window for application", app_id)
        for w in self._get_wnck_screen_windows_stacked():
            app = w.get_application()
            app_pid = app.get_pid()
            if not app_pid:
                app_pid = w.get_pid()
            if app_pid == pid:
                self._store(app_id, w)
                return False
        if time() > timeout:
            return False
        return True

    def application_name(self, app_id):
        if not self._has_match(app_id):
            return None
        return self.register[app_id]

    def application_is_running(self, app_id):
        for w in self._get_wnck_screen_windows_stacked():
            if (w.get_application() and self._is_match(app_id, w) and 
                w.get_window_type() == Wnck.WindowType.NORMAL):
                return True
        return False

    def get_application_windows(self, app_id):
        application_windows = []
        for w in self._get_wnck_screen_windows_stacked():
            if (w.get_application() and self._is_match(app_id, w) and 
                w.get_window_type() == Wnck.WindowType.NORMAL):
                application_windows.append(w)
        return application_windows

    def application_to_front(self, app_id):
        application_windows = self.get_application_windows(app_id)
        if not application_windows:
            return False
        etime = uievents.current_event_time()
        # if True, focus app's all windows on the same workspace
        # if False, focus only one window (in cyclical manner)
        focus_all = True
        if focus_all:
            return self._to_front_application_style(application_windows, etime)
        else:
            return self._to_front_single(application_windows, etime)

    def _to_front_application_style(self, application_windows, evttime):
        workspaces = {}
        cur_screen = application_windows[0].get_screen()
        cur_workspace = cur_screen.get_active_workspace()

        def visible_window(window):
            return (window.get_window_type() == Wnck.WindowType.NORMAL and
                    window.is_visible_on_workspace(cur_workspace))

        def normal_window(window):
            return window.get_window_type() == Wnck.WindowType.NORMAL

        ## get all visible windows in stacking order
        vis_windows = list(filter(visible_window,
                             self._get_wnck_screen_windows_stacked()))

        ## sort windows into "bins" by workspace
        for w in filter(normal_window, application_windows):
            wspc = w.get_workspace() or cur_workspace
            workspaces.setdefault(wspc, []).append(w)

        cur_wspc_windows = workspaces.get(cur_workspace, [])
        # make a rotated workspace list, with current workspace first
        idx = cur_workspace.get_number()
        all_workspaces = cur_screen.get_workspaces()
        all_workspaces[:] = all_workspaces[idx:] + all_workspaces[:idx]
        # check if the application's window on current workspace
        # are the topmost
        focus_windows = []
        if (cur_wspc_windows and 
            set(vis_windows[-len(cur_wspc_windows):]) != set(cur_wspc_windows)):
            focus_windows = cur_wspc_windows
            ## if the topmost window is already active, take another
            if focus_windows[-1:] == vis_windows[-1:]:
                focus_windows[:] = focus_windows[:-1]
        else:
            # all windows are focused, find on next workspace
            for wspc in all_workspaces[1:]:
                focus_windows = workspaces.get(wspc, [])
                if focus_windows:
                    break
            else:
                # no windows on other workspaces, so we rotate among
                # the local ones
                focus_windows = cur_wspc_windows[:1]
        self._focus_windows(focus_windows, evttime)

    def _to_front_single(self, application_windows, evttime):
        # bring the first window to front
        for window in application_windows:
            self._focus_windows([window], evttime)
            return

    def _focus_windows(self, windows, evttime):
        for window in windows:
            # we special-case the desktop
            # only show desktop if it's the only window
            if window.get_name() == "x-nautilus-desktop":
                if len(windows) == 1:
                    screen = Wnck.Screen.get_default()
                    if screen is not None:
                        screen.toggle_showing_desktop(True)
                else:
                    continue
            wspc = window.get_workspace()
            if wspc:
                wspc.activate(evttime)
            window.activate_transient(evttime)

    def application_close_all(self, app_id):
        application_windows = self.get_application_windows(app_id)
        evttime = uievents.current_event_time()
        for w in application_windows:
            if not w.is_skip_tasklist():
                w.close(evttime)


_appl_match_service = None
def GetApplicationsMatcherService():
    """Get the (singleton) ApplicationsMatcherService"""
    global _appl_match_service
    if not _appl_match_service:
        _appl_match_service = ApplicationsMatcherService()
    return _appl_match_service




__version__ = '2010-01-21'


from kupfer.obj.objects import RunnableLeaf
from kupfer import kupferui


class PleaseConfigureLeaf(RunnableLeaf):
    """ Show information and allow to open preferences for given plugin """
    message = _("Please Configure Plugin")
    description = _("Plugin %s is not configured")

    def __init__(self, plugin_id, plugin_name):
        plugin_id = plugin_id.split('.')[-1]
        RunnableLeaf.__init__(self, plugin_id, self.message)
        self.plugin_name = plugin_name

    def wants_context(self):
        return True

    def run(self, ctx):
        kupferui.show_plugin_info(self.object, ctx.environment)

    def get_icon_name(self):
        return "preferences-desktop"

    def get_description(self):
        return self.description % self.plugin_name


class InvalidCredentialsLeaf(PleaseConfigureLeaf):
    description = _("Invalid user credentials for %s")





# -*- coding: UTF-8 -*-
'''
virtualbox_ose_support.py

Control VirtualBox via command-line interface.
Support both Sun VirtualBox and VirtualBox OpenSource Edition.
'''

__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = '2018-10-21'

import os
from xml.dom import minidom

from kupfer import pretty, utils
from kupfer.plugin.virtualbox import constants as vbox_const

_VBOX_CONFIG_DIRS = (
    os.path.expanduser('~/.config/VirtualBox/'),
    os.path.expanduser('~/.VirtualBox/')
)
_VBOX_CONFIG_FILE = 'VirtualBox.xml'

MONITORED_DIRS = _VBOX_CONFIG_DIRS
IS_DYNAMIC = False
ICON = "virtualbox-ose"
APP_ID = "virtualbox-ose"

# parameters for VBoxManage
_ACTIONS = {
    vbox_const.VM_POWEROFF: 'poweroff',
    vbox_const.VM_ACPI_POWEROFF: 'acpipowerbutton',
    vbox_const.VM_PAUSE: 'pause',
    vbox_const.VM_REBOOT: 'reset',
    vbox_const.VM_RESUME: 'resume',
    vbox_const.VM_SAVE: 'savestate',
}


def get_machine_state(vm_uuid):
    ''' check vms state (on/off/paused) '''
    state = vbox_const.VM_STATE_POWEROFF
    try:
        str_state = 'poweroff'
        with os.popen('VBoxManage showvminfo %s --machinereadable' % vm_uuid) \
                as pinfo:
            for line in pinfo:
                if line.startswith('VMState="'):
                    str_state = line.strip()[9:-1]
                    break
        if str_state == 'paused':
            state = vbox_const.VM_STATE_PAUSED
        elif str_state == 'running':
            state = vbox_const.VM_STATE_POWERON
        elif str_state == 'saved':
            state = vbox_const.VM_STATE_SAVED
    except IOError as err:
        pretty.print_error(__name__, 'get_machine_state', vm_uuid,
                           'error', err)
        state = vbox_const.VM_STATE_POWEROFF

    return state


def vm_action(action, vm_uuid):
    ''' change state of the virtual machine. Call VBoxManage.
        @param action - one of the const VM_*
        @param vm_uuid - virtual machine uuid
    '''
    if action == vbox_const.VM_START_NORMAL:
        utils.spawn_async(['VBoxManage', 'startvm', vm_uuid, '--type', 'gui'])
    elif action == vbox_const.VM_START_HEADLESS:
        utils.spawn_async(['VBoxManage', 'startvm', vm_uuid, '--type',
                           'headless'])
    else:
        command = _ACTIONS[action]
        utils.spawn_async(['VBoxManage', 'controlvm', vm_uuid, command])


def _get_virtual_machines(config_file):
    ''' load (virtual machine uuid, path to vm config) from virtualbox
        configuration.
        @param config_file - path to VirtualBox.xml file
    '''
    try:
        dtree = minidom.parse(config_file)
        machine_registry = dtree.getElementsByTagName('MachineRegistry')[0]
        print(machine_registry)
        for machine in machine_registry.getElementsByTagName('MachineEntry'):
            yield (machine.getAttribute('uuid')[1:-1],
                   machine.getAttribute('src'))
    except Exception as err:
        pretty.print_error(__name__, '_get_virtual_machines', config_file,
                           'error', err)


def _get_machine_info(vm_uuid, config_file):
    ''' load information about virtual machines from its configuration file.
        @param vm_uuid - uuid virtual machine
        @param config_file - path to vm configuration file
    '''
    if not os.path.isfile(config_file):
        return None, None

    try:
        dtree = minidom.parse(config_file)
        machine_registry = dtree.getElementsByTagName('Machine')[0]
        os_type = machine_registry.getAttribute('OSType')
        name = machine_registry.getAttribute('name')
        description = None
        for machine_registry_child in machine_registry.childNodes:
            if machine_registry_child.nodeName == 'Description':
                if machine_registry_child.hasChildNodes():
                    description = machine_registry_child.firstChild.nodeValue
                break
        return (name, description or os_type)
    except Exception as err:
        pretty.print_error(__name__, '_get_machine_info', vm_uuid, 'error' +
                           config_file, err)
    return None, None


def get_machines():
    for dirname in _VBOX_CONFIG_DIRS:
        vbox_conf = os.path.join(dirname, _VBOX_CONFIG_FILE)
        if not os.path.isfile(vbox_conf):
            continue
        for vm_uuid, config in _get_virtual_machines(vbox_conf):
            if not os.path.isabs(config):
                config = os.path.join(dirname, config)
            name, description = _get_machine_info(vm_uuid, config)
            if name:
                yield (vm_uuid, name, description)


def unload():
    pass
